#!/usr/bin/env python
from PyQt5 import QtWidgets, QtCore, QtGui
import os
import shutil
import PyDP4
import queue
import sys
from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure
import numpy as np
from scipy.stats import norm
from PyQt5.QtSvg import QSvgWidget
from rdkit.Chem import AllChem as Chem
from rdkit.Chem.Draw import rdMolDraw2D
import pickle
from matplotlib import cm

from pathlib import Path

class Window(QtWidgets.QMainWindow):
    # create signal to start background job

    signal_start_PyDP4 = QtCore.pyqtSignal()

    # signal_finished_PyDP4 = QtCore.pyqtSignal()

    def __init__(self):
        super(Window, self).__init__()

        self.table_widget = TabWidget(self)
        self.setCentralWidget(self.table_widget)

        self.resize(self.table_widget.Tab1.Overall_widget.sizeHint())
        #self.resize(self.table_widget.sizeHint())
        self.show()

        # check if DP4outfile is there


class TabWidget(QtWidgets.QWidget):

    def __init__(self, parent):
        super(QtWidgets.QWidget, self).__init__(parent)
        self.layouttabs = QtWidgets.QVBoxLayout()

        # Initialize tab screen
        self.tabs = QtWidgets.QTabWidget()

        self.tab1 = QtWidgets.QWidget()

        self.Tab1 = CalculationTab()

        self.tabs.addTab(self.Tab1, "Calculation")
        self.tab1.layout = QtWidgets.QVBoxLayout()

        # Add tabs to widget
        self.layouttabs.addWidget(self.tabs)

        self.setLayout(self.layouttabs)

    def addplottabs(self):

        self.tab2 = QtWidgets.QWidget()
        self.tab3 = QtWidgets.QWidget()
        self.tab4 = QtWidgets.QWidget()
        self.tab5 = QtWidgets.QWidget()
        self.tab6 = QtWidgets.QWidget()

        self.tabs.addTab(self.tab5, "Conformers")

        #only add proton/carbon plot tabs if the raw data is there

        NMR_list = [str(self.Tab1.NMR_list.item(i).text()) for i in range(self.Tab1.NMR_list.count())]

        if 'Proton' in NMR_list or 'Proton.dx' in NMR_list or 'proton' in NMR_list or 'proton.dx' in NMR_list:

            self.tabs.addTab(self.tab2, "Proton Plot")
            self.Tab2 = ProtonPlotTab()
            self.tab2.layout = QtWidgets.QVBoxLayout(self)
            self.tab2.layout.addWidget(self.Tab2)
            self.tab2.setLayout(self.tab2.layout)

        if 'Carbon' in NMR_list or 'Carbon.dx' in NMR_list or 'carbon' in NMR_list or 'carbon.dx' in NMR_list:

            self.tabs.addTab(self.tab3, "Carbon Plot")
            self.Tab3 = CarbonPlotTab()
            self.tab3.layout = QtWidgets.QVBoxLayout(self)
            self.tab3.layout.addWidget(self.Tab3)
            self.tab3.setLayout(self.tab3.layout)

        self.Tab5 = ConformerTab()
        self.tab5.layout = QtWidgets.QVBoxLayout(self)
        self.tab5.layout.addWidget(self.Tab5)
        self.tab5.setLayout(self.tab5.layout)

        self.Tab6 = DP5Tab()

        if 's' in ui.table_widget.Tab1.settings.Workflow:
            self.tabs.addTab(self.tab6, "DP5")
            self.tab6.layout = QtWidgets.QVBoxLayout(self)
            self.tab6.layout.addWidget(self.Tab6)
            self.tab6.setLayout(self.tab6.layout)

        self.Tab4 = StatsTab()
        if 's' in ui.table_widget.Tab1.settings.Workflow and len(ui.table_widget.Tab1.settings.InputFiles) > 1:
            self.tabs.addTab(self.tab4, "DP4")
            self.tab4.layout = QtWidgets.QVBoxLayout(self)
            self.tab4.layout.addWidget(self.Tab4)
            self.tab4.setLayout(self.tab4.layout)




        self.layouttabs.addWidget(self.tabs)
        self.setLayout(self.layouttabs)

        ui.update()


class DP5Tab(QtWidgets.QWidget):

    def __init__(self):
        super(DP5Tab, self).__init__()

        # main horizontal widget

        self.main_layout = QtWidgets.QHBoxLayout()
        self.setLayout(self.main_layout)
        self.Isomers = ui.table_widget.Tab1.worker.Isomers
        self.DP5data = ui.table_widget.Tab1.worker.DP5Data
        self.DP4data = ui.table_widget.Tab1.worker.DP4Data

        self.Isomer_DP5_table = QtWidgets.QTableWidget(self)

        self.Isomer_DP5_table.setColumnCount(3)

        self.Isomer_DP5_table.setHorizontalHeaderLabels(["Isomer", "DP5", "DP4"])

        self.Isomer_DP5_table.setRowCount(len(self.DP5data.DP5scaledprobs))

        c = 0

        for scaled_DP5, DP4 in zip(self.DP5data.DP5scaledprobs, self.DP4data.DP4probs):

            self.Isomer_DP5_table.setItem(c, 0, QtWidgets.QTableWidgetItem(str(c + 1)))

            self.Isomer_DP5_table.setItem(c, 1, QtWidgets.QTableWidgetItem(str(round(scaled_DP5, 2))))

            self.Isomer_DP5_table.setItem(c, 2, QtWidgets.QTableWidgetItem(str(round(DP4 , 3))))

            c += 1

        self.Isomer_DP5_table.selectRow(0)

        # add error table

        self.DP5_table = QtWidgets.QTableWidget(self)

        self.DP5_table.setColumnCount(4)

        self.DP5_table.setHorizontalHeaderLabels(
            ["Atom Label", "Atomic Probability",  "Scaled error (ppm)", "Unscaled error (ppm)"])

        self.DP5_table.setColumnWidth(0, 70)
        self.DP5_table.setColumnWidth(1, 70)
        self.DP5_table.setColumnWidth(2, 70)
        self.DP5_table.setColumnWidth(3, 70)

        #########

        # add isomer canvas

        self.table_widget = QtWidgets.QWidget(self)

        self.image = QSvgWidget()

        self.image.setMinimumWidth(self.image.height())

        # organise layouts

        self.table_layout = QtWidgets.QVBoxLayout()

        self.table_layout.addWidget(self.Isomer_DP5_table)

        self.table_layout.addWidget(self.DP5_table)

        #self.image_layout.addWidget(self.image)

        self.table_widget.setLayout(self.table_layout)

        self.main_layout.addWidget(self.table_widget)

        self.main_layout.addWidget(self.image)

        # connections

        self.Isomer_DP5_table.currentCellChanged.connect(self.Mol_drawer)

        self.Isomer_DP5_table.currentCellChanged.connect(self.populatetable)

        self.DP5_table.currentCellChanged.connect(self.Mol_drawer)

    def populatetable(self):

        isomerindex = self.Isomer_DP5_table.currentRow()

        # which isomer is selected

        self.DP5_table.setRowCount(0)

        self.labels, self.shifts, self.exps, self.scaleds = self.DP5data.Clabels[isomerindex], \
                                                            self.DP5data.Cshifts[isomerindex], \
                                                            self.DP5data.Cexp[isomerindex], \
                                                            self.DP5data.Cscaled[isomerindex], \

        self.scaled_probs = self.DP5data.BScaledAtomProbs[isomerindex]

        self.DP5_table.setRowCount(len(self.labels))

        # set info in rows and columns

        c = 0

        for label, shift, exp, scaled_shift,  scaled_prob in zip(self.labels, self.shifts, self.exps,
                                                                               self.scaleds,
                                                                               self.scaled_probs):
            self.DP5_table.setItem(c, 0, QtWidgets.QTableWidgetItem(label))
            self.DP5_table.setItem(c, 1, QtWidgets.QTableWidgetItem(str(round(scaled_prob, 2))))

            self.DP5_table.setItem(c, 2, QtWidgets.QTableWidgetItem(str(round(abs(scaled_shift - exp), 2))))
            self.DP5_table.setItem(c, 3, QtWidgets.QTableWidgetItem(str(round(abs(shift - exp), 2))))

            c += 1

        self.DP5_table.selectRow(0)

        # self.plot()

    def Isomer_number(self):
        Isomer_list = []

        for c, i in enumerate(self.Isomers):
            Isomer_list.append("Isomer " + str(c + 1))

        return Isomer_list

    def Mol_drawer(self):

        isomerindex = self.Isomer_DP5_table.currentRow()

        highlight = {}

        inds = []

        for label in self.DP5data.Clabels[isomerindex]:
            inds.append(int(label.split("C")[1]) - 1)

        for i, s in zip(inds, self.DP5data.BScaledAtomProbs[isomerindex]):

            highlight[i] = cm.RdYlGn(s)

        # pick selected atom

        if self.DP5_table.item(self.DP5_table.currentRow(), 0):

            error_ind = int(self.DP5_table.currentRow())

            atom_ind = inds[error_ind]

            highlight[atom_ind] = cm.viridis(  self.DP5data.BScaledAtomProbs[isomerindex][error_ind])

        #m = Chem.MolFromMolFile(str(ui.table_widget.Tab1.worker.settings.InputFiles[0]).split('.sdf')[0] + '.sdf', removeHs=False)

        m = Chem.MolFromMolFile(str(ui.table_widget.Tab1.worker.settings.InputFilesPaths[isomerindex]).split('.sdf')[0] + '.sdf',removeHs=False)

        Chem.Compute2DCoords(m)

        drawer = rdMolDraw2D.MolDraw2DSVG(self.image.width(), self.image.height())

        drawer.DrawMolecule(m, highlightAtoms=inds, highlightAtomColors=highlight, highlightBonds=False)

        # ,addStereoAnnotation = True,addAtomIndices = True

        drawer.FinishDrawing()

        svg = drawer.GetDrawingText().replace('svg:', '')

        svg_bytes = bytearray(svg, encoding='utf-8')

        self.image.renderer().load(svg_bytes)

        ui.update()


class StatsTab(QtWidgets.QWidget):

    def __init__(self):
        super(StatsTab, self).__init__()

        #main horizontal widget

        self.main_layout = QtWidgets.QHBoxLayout()
        self.setLayout(self.main_layout)
        self.Isomers = ui.table_widget.Tab1.worker.Isomers
        self.dp4data = ui.table_widget.Tab1.worker.DP4Data

        # add isomer selection
        self.IsomerSelect = QtWidgets.QComboBox(self)
        self.IsomerSelect.addItems(self.Isomer_number())

        # add isomer canvas

        self.image = QSvgWidget()

        #make vertical isomer layout

        self.isomer_layout = QtWidgets.QVBoxLayout()
        self.isomer_layout.addWidget(self.IsomerSelect)
        self.isomer_layout.addWidget(self.image)

        self.isomer_widget = QtWidgets.QWidget(self)
        self.isomer_widget.setLayout(self.isomer_layout)

        # make error table

        # populate table based on the Isomer selected

        self.Hplot = plotstats('H')

        #check if there is a DP4 result for this atomt type:

        if not np.isnan(ui.table_widget.Tab1.worker.DP4Data.HDP4probs).any() :

            self.IsomerSelect.currentIndexChanged.connect(self.Hplot.populatetable)

            self.IsomerSelect.currentIndexChanged.connect(self.RenderImage)

            self.Hplot.errortable.itemSelectionChanged.connect(self.RenderImage)

        self.Cplot = plotstats('C')

        if not np.isnan(ui.table_widget.Tab1.worker.DP4Data.CDP4probs).any() :

            self.IsomerSelect.currentIndexChanged.connect(self.Cplot.populatetable)

            self.IsomerSelect.currentIndexChanged.connect(self.RenderImage)

            self.Cplot.errortable.itemSelectionChanged.connect(self.RenderImage)

        #make error table layout

        self.error_layout = QtWidgets.QVBoxLayout()

        self.error_layout.addWidget(self.Cplot.errortable)
        self.error_layout.addWidget(self.Hplot.errortable)

        self.error_widget = QtWidgets.QWidget(self)
        self.error_widget.setLayout(self.error_layout)

        #make plot layout

        self.plot_layout = QtWidgets.QVBoxLayout()

        self.plot_layout.addWidget(self.Cplot.statscanvas)
        self.plot_layout.addWidget(self.Hplot.statscanvas)

        self.plot_widget = QtWidgets.QWidget(self)
        self.plot_widget.setLayout(self.plot_layout)

        # add these widgets to main horizontal layout

        self.isomer_widget.setMinimumWidth(400)

        self.error_widget.setFixedWidth(400)

        self.main_layout.addWidget(self.isomer_widget)
        self.main_layout.addWidget(self.error_widget)
        self.main_layout.addWidget(self.plot_widget)

    def Isomer_number(self):
        Isomer_list = []

        for c, i in enumerate(self.Isomers):
            Isomer_list.append("Isomer " + str(c + 1))

        return Isomer_list

    def RenderImage(self):

        '''colors = [(0.12, 0.47, 0.71), (1.0, 0.5, 0.05), (0.17, 0.63, 0.17), (0.84, 0.15, 0.16), (0.58, 0.4, 0.74),
                  (0.55, 0.34, 0.29), (0.89, 0.47, 0.76), (0.5, 0.5, 0.5), (0.74, 0.74, 0.13), (0.09, 0.75, 0.81)]'''



        isomerindex = self.IsomerSelect.currentIndex()

        atom = []

        highlight = {}

        if self.Cplot.errortable.item(self.Cplot.errortable.currentRow(), 0):

            C_atom = int(self.Cplot.errortable.item(self.Cplot.errortable.currentRow(), 0).text().split("C")[1])

            atom.append(C_atom - 1 )

            highlight[C_atom - 1] = (0.12, 0.47, 0.71)

        if self.Hplot.errortable.item(self.Hplot.errortable.currentRow(), 0):

            H_atom = int(self.Hplot.errortable.item(self.Hplot.errortable.currentRow(), 0).text().split("H")[1])

            atom.append(H_atom - 1)

            highlight[H_atom - 1] = (0.84, 0.15, 0.16)

        m = Chem.MolFromMolFile(str(ui.table_widget.Tab1.worker.settings.InputFilesPaths[isomerindex]).split('.sdf')[0] + '.sdf',removeHs=False)
        Chem.Compute2DCoords(m)

        drawer = rdMolDraw2D.MolDraw2DSVG(self.image.width(), self.image.height())

        drawer.DrawMolecule(m, highlightAtoms=atom, highlightAtomColors=highlight,highlightBonds = False)

        drawer.FinishDrawing()

        svg = drawer.GetDrawingText().replace('svg:', '')

        svg_bytes = bytearray(svg, encoding='utf-8')

        self.image.renderer().load(svg_bytes)

        #self.image.setGeometry(QtCore.QRect(0, 50, 500, 500))

        ui.update()


class plotstats(QtWidgets.QWidget):

    def __init__(self, Atom):

        super(plotstats, self).__init__()

        # when a cell is selected signal to plot stats

        self.atom = Atom

        self.findmeans()

        self.dp4data = ui.table_widget.Tab1.worker.DP4Data

        self.Isomers = ui.table_widget.Tab1.worker.Isomers

        self.errortable = QtWidgets.QTableWidget(self)

        self.errortable.setColumnCount(5)

        self.errortable.setHorizontalHeaderLabels(["Atom Label", "Calc Shift", "Scaled", "Exp", "Error"])

        self.errortable.setColumnWidth(0, 70)
        self.errortable.setColumnWidth(1, 70)
        self.errortable.setColumnWidth(2, 70)
        self.errortable.setColumnWidth(3, 70)
        self.errortable.setColumnWidth(4, 70)

        self.statsfigure = Figure()

        self.statscanvas = FigureCanvas(self.statsfigure)

        # self.errortable.cellClicked.connect(self.plot)

        self.errortable.itemSelectionChanged.connect(self.plot)

        self.statscanvas.mpl_connect('button_press_event', self.selectpoint)

    def plot(self):

        r = int(self.errortable.currentRow())

        if r >= 0:

            error = float(self.errortable.item(r, 4).text())

            # decide which stats params have been used

            self.statsfigure.clear()

            self.statsfig = self.statsfigure.add_subplot(111)

            if self.atom == 'H':

                m = abs(max([item for sublist in self.dp4data.Herrors for item in sublist], key=abs))

                p_color = 3

            elif self.atom == 'C':

                m = abs(max([item for sublist in self.dp4data.Cerrors for item in sublist], key=abs))

                p_color = 0

            # plot all errors at low transparency

            for e in self.errors:

                self.statsfig.plot([e, e], [0, self.multipdf([float(e)])], color='C'+ str(p_color), alpha=0.25)

                self.statsfig.plot(e, self.multipdf([float(e)]), 'o', color='C'+ str(p_color), alpha=0.25)

            x = np.linspace(- 2 * m, 2 * m, 1000)

            self.statsfig.plot(x, self.multipdf(x),color = "C1")


            self.statsfig.plot([error, error], [0, self.multipdf([float(error)])], color='C'+ str(p_color), alpha=1)

            self.statsfig.plot(error, self.multipdf([float(error)]), 'o', color='C'+ str(p_color), alpha= 1)


            self.statsfig.set_xlim([x[0], x[-1]])

            self.statsfig.set_xlabel("error (ppm)")

            self.statsfig.set_ylabel("probability density")

            self.statscanvas.draw()

    def multipdf(self, x):

        y = np.zeros(len(x))

        if self.atom == 'H':

            for mean, std in zip(self.Hmeans, self.Hstdevs):
                y += norm.pdf(x, loc=mean, scale=std)

        elif self.atom == 'C':

            for mean, std in zip(self.Cmeans, self.Cstdevs):
                y += norm.pdf(x, loc=mean, scale=std)

        return y

    def findmeans(self):

        if ui.table_widget.Tab1.worker.settings.StatsParamFile == 'none':

            self.Hmeans = [0]

            self.Hstdevs = [0.18731058105269952]

            self.Cmeans = [0]

            self.Cstdevs = [2.269372270818724]

        else:

            self.Cmeans, self.Cstdevs, self.Hmeans, self.Hstdevs = ReadParamFile(
                ui.table_widget.Tab1.worker.settings.StatsParamFile, 'g')

    def populatetable(self):

        # which isomer is selected

        self.isomerindex = int(str(ui.table_widget.Tab4.IsomerSelect.currentText())[-1]) - 1

        self.errortable.setRowCount(0)

        if self.atom == 'H':

            self.labels, self.shifts, self.exps, self.scaleds, self.errors = self.dp4data.Hlabels[self.isomerindex], \
                                                                             self.dp4data.Hshifts[self.isomerindex], \
                                                                             self.dp4data.Hexp[self.isomerindex], \
                                                                             self.dp4data.Hscaled[self.isomerindex], \
                                                                             self.dp4data.Herrors[self.isomerindex]

        elif self.atom == 'C':

            self.labels, self.shifts, self.exps, self.scaleds, self.errors = self.dp4data.Clabels[self.isomerindex], \
                                                                             self.dp4data.Cshifts[self.isomerindex], \
                                                                             self.dp4data.Cexp[self.isomerindex], \
                                                                             self.dp4data.Cscaled[self.isomerindex], \
                                                                             self.dp4data.Cerrors[self.isomerindex]

        self.errortable.setRowCount(len(self.labels))

        # set info in rows and columns

        c = 0

        for label, shift, exp, scaled, error in zip(self.labels, self.shifts, self.exps, self.scaleds, self.errors):
            self.errortable.setItem(c, 0, QtWidgets.QTableWidgetItem(label))
            self.errortable.setItem(c, 1, QtWidgets.QTableWidgetItem(str(round(shift, 2))))
            self.errortable.setItem(c, 2, QtWidgets.QTableWidgetItem(str(round(scaled, 2))))
            self.errortable.setItem(c, 3, QtWidgets.QTableWidgetItem(str(round(exp, 2))))
            self.errortable.setItem(c, 4, QtWidgets.QTableWidgetItem(str(round(error, 2))))

            c += 1

        self.errortable.selectRow(0)

        self.plot()

    def selectpoint(self, event):

        self.xpos = event.xdata

        self.ypos = event.ydata

        # find the closest point to the click

        coords = self.statsfig.transData.transform((self.xpos, self.ypos))

        coordinates = np.array(self.statsfig.transData.transform(list(zip(self.errors, self.multipdf(self.errors)))))

        mindis = np.argmin((coordinates[:, 0] - coords[0]) ** 2 + (coordinates[:, 1] - coords[1]) ** 2)

        self.errortable.selectRow(mindis)


class CalculationTab(QtWidgets.QWidget):

    signal_start_PyDP4 = QtCore.pyqtSignal()

    def __init__(self):
        self.log_file = ''
        super(CalculationTab, self).__init__()
        self.cwd = Path(os.getcwd())

        self.Overall_verticallayout =QtWidgets.QVBoxLayout(self)
        self.Overall_widget = QtWidgets.QWidget(self)

        self.title = QtWidgets.QLabel(self)
        self.title.setText("DP5 GUI")
        self.title.setObjectName("Title")
        self.Overall_verticallayout.addWidget(self.title)

        ########### input

        self.input_widget = QtWidgets.QWidget(self)
        self.input_layout = QtWidgets.QHBoxLayout(self)

        self.structure_widget = QtWidgets.QWidget(self)
        self.structure_layout = QtWidgets.QVBoxLayout(self)

        self.Add_structure = QtWidgets.QPushButton(self)
        self.Add_structure.setObjectName("Add_structure")
        self.Add_structure.setText("Add structure")

        self.remove_structure = QtWidgets.QPushButton(self)
        self.remove_structure.setObjectName("remove_structure")
        self.remove_structure.setText("Remove selected")

        self.structure_layout.addWidget(self.Add_structure)
        self.structure_layout.addWidget(self.remove_structure)

        self.structure_widget.setLayout(self.structure_layout)

        self.input_layout.addWidget(self.structure_widget)

        self.Structure_list = QtWidgets.QListWidget(self)
        self.Structure_list.setMaximumHeight(self.Add_structure.sizeHint().height() * 2)

        self.Structure_list.setObjectName("Structure_list")
        self.input_layout.addWidget(self.Structure_list)

        self.NMR_widget = QtWidgets.QWidget(self)
        self.NMR_layout = QtWidgets.QVBoxLayout(self)

        self.Add_NMR = QtWidgets.QPushButton(self)
        self.Add_NMR.setObjectName("Add_NMR")
        self.Add_NMR.setText("Add raw NMR")

        self.Add_NMR_desc = QtWidgets.QPushButton(self)
        self.Add_NMR_desc.setObjectName("Add_NMR")
        self.Add_NMR_desc.setText("Add NMR desc")

        self.remove_NMR = QtWidgets.QPushButton(self)
        self.remove_NMR.setObjectName("remove_NMR")
        self.remove_NMR.setText("Remove selected")

        self.NMR_layout.addWidget(self.Add_NMR)
        self.NMR_layout.addWidget(self.Add_NMR_desc)
        self.NMR_layout.addWidget(self.remove_NMR)

        self.NMR_widget.setLayout(self.NMR_layout)
        self.input_layout.addWidget(self.NMR_widget)

        self.NMR_list = QtWidgets.QListWidget(self)
        self.NMR_list.setObjectName("NMR_list")
        self.NMR_list.setMaximumHeight(self.Add_structure.sizeHint().height() * 3)
        self.input_layout.addWidget(self.NMR_list)

        self.Output_add = QtWidgets.QPushButton(self)
        self.Output_add.setObjectName("Out add")
        self.Output_add.setText("Output Folder")
        self.input_layout.addWidget(self.Output_add)

        self.Output_list = QtWidgets.QListWidget(self)
        self.Output_list.setObjectName("Out list")
        self.Output_list.setMaximumHeight(self.Add_structure.sizeHint().height() * 2)
        self.input_layout.addWidget(self.Output_list)

        self.input_widget.setLayout(self.input_layout)
        self.Overall_verticallayout.addWidget(self.input_widget)

        ############ workflow

        self.workflow_layout = QtWidgets.QGridLayout(self)
        self.workflow_widget = QtWidgets.QWidget(self)

        self.workflow_title = QtWidgets.QLabel(self)
        self.workflow_title.setObjectName("workflow_title")
        self.workflow_title.setText("Workflow")
        self.workflow_layout.addWidget(self.workflow_title, 0, 0)

        self.CleanUp_yn = QtWidgets.QCheckBox(self)
        self.CleanUp_yn.setObjectName("CleanUp_yn")
        self.CleanUp_yn.setText("Clean\n"
                                          "Structures")

        self.workflow_layout.addWidget(self.CleanUp_yn,1,0)

        self.Gen_diastereomers_yn = QtWidgets.QCheckBox(self)
        self.Gen_diastereomers_yn.setObjectName("Gen_diastereomers_yn")
        self.Gen_diastereomers_yn.setText("Generate\n"
                                          "Diastereomers")

        self.workflow_layout.addWidget(self.Gen_diastereomers_yn,1,1)

        self.Solvent_yn = QtWidgets.QCheckBox(self)
        self.Solvent_yn.setObjectName("Solvent_yn")
        self.Solvent_yn.setText("Solvent")
        self.workflow_layout.addWidget(self.Solvent_yn,1,2)

        self.solvent_drop = QtWidgets.QComboBox(self)
        self.solvent_drop.setObjectName("solvent_drop")
        solvents = ['chloroform', 'dimethylsulfoxide', 'benzene', 'methanol', 'pyridine', 'acetone']
        self.solvent_drop.addItems(solvents)

        self.workflow_layout.addWidget(self.solvent_drop,2,2)

        self.MM_yn = QtWidgets.QCheckBox(self)
        self.MM_yn.setObjectName("MM_yn")
        self.MM_yn.setText("Molecular\nMechanics")
        self.workflow_layout.addWidget(self.MM_yn,1,3)

        self.MM_drop = QtWidgets.QComboBox(self)
        self.MM_drop.setObjectName("MM_drop")
        MM = ["MacroModel","Tinker"]
        self.MM_drop.addItems(MM)
        self.workflow_layout.addWidget(self.MM_drop,2,3)

        self.MM_advanced = QtWidgets.QPushButton(self)
        self.MM_advanced.setText("Advanced Settings")
        self.workflow_layout.addWidget(self.MM_advanced,3,3)

        self.DFT_yn = QtWidgets.QCheckBox(self)
        self.DFT_yn.setObjectName("DFT_yn")
        self.DFT_yn.setText("DFT\nCalculations")
        self.workflow_layout.addWidget(self.DFT_yn, 1, 4)

        self.DFT_drop = QtWidgets.QComboBox(self)
        self.DFT_drop.setObjectName("DFT_drop")
        DFT = ["Gaussian", "NWChem"]
        self.DFT_drop.addItems(DFT)
        self.workflow_layout.addWidget(self.DFT_drop, 2, 4)

        self.DFT_advanced = QtWidgets.QPushButton(self)
        self.DFT_advanced.setText("Advanced Settings")
        self.workflow_layout.addWidget(self.DFT_advanced,3,4)

        self.Assignment_yn = QtWidgets.QCheckBox(self)
        self.Assignment_yn.setObjectName("Assignment_yn")
        self.Assignment_yn.setText("NMR\nAssignment")
        self.workflow_layout.addWidget(self.Assignment_yn, 1, 5)

        self.DP5_stat_yn = QtWidgets.QCheckBox(self)
        self.DP5_stat_yn.setObjectName("DP5_stat_yn")
        self.DP5_stat_yn.setText("DP5 Statistics")
        self.workflow_layout.addWidget(self.DP5_stat_yn, 1, 6)

        self.DP4_stat_yn = QtWidgets.QCheckBox(self)
        self.DP4_stat_yn.setObjectName("DP4_stat_yn")
        self.DP4_stat_yn.setText("DP4 Statistics")
        self.workflow_layout.addWidget(self.DP4_stat_yn, 1, 7)

        self.Add_stats_model = QtWidgets.QPushButton(self)
        self.Add_stats_model.setObjectName("Add_stats_model")
        self.Add_stats_model.setText("Add stats model")
        self.workflow_layout.addWidget(self.Add_stats_model, 2, 7)

        self.Stats_list = QtWidgets.QListWidget(self)
        self.Stats_list.setMaximumHeight(self.Add_structure.sizeHint().height())
        self.Stats_list.setObjectName("Stats_list")

        self.workflow_layout.addWidget(self.Stats_list, 3, 7)
        self.workflow_widget.setLayout(self.workflow_layout)
        self.Overall_verticallayout.addWidget(self.workflow_widget)

        ############ calc

        self.Gobutton = QtWidgets.QPushButton(self)
        self.Gobutton.setObjectName("Gobutton")
        self.Gobutton.setText("Calculate")
        self.Overall_verticallayout.addWidget(self.Gobutton)

        self.Output_box = QtWidgets.QTextEdit(self)
        self.Output_box.setObjectName("Output_box")
        self.Overall_verticallayout.addWidget(self.Output_box)

        self.Overall_widget.setLayout(self.Overall_verticallayout)

        #################################################################################################buttons methods

        # adding structures to the gui

        self.NMR_paths = []

        self.Structure_paths = []

        self.Output_folder = Path.cwd()

        self.Add_structure.clicked.connect(self.addstructure)

        self.remove_structure.clicked.connect(self.removestructure)

        self.Add_NMR.clicked.connect(self.addNMR)

        self.Add_NMR_desc.clicked.connect(self.remove_all_NMR)

        self.Add_NMR_desc.clicked.connect(self.addNMRdesc)

        self.remove_NMR.clicked.connect(self.removeNMR)

        self.Output_add.clicked.connect(self.addoutputfolder)

        # selecting solvent check box

        self.solvent_drop.setEnabled(False)

        self.Solvent_yn.stateChanged.connect(self.solventtoggle)

        self.MM_drop.setEnabled(False)
        self.MM_advanced.setEnabled(False)

        self.MM_yn.stateChanged.connect(self.MMtoggle)

        self.DFT_drop.setEnabled(False)
        self.DFT_advanced.setEnabled(False)

        self.DFT_yn.stateChanged.connect(self.DFTtoggle)

        self.Assignment_yn.stateChanged.connect(self.Assignment_toggle)

        # adding a stats model

        self.Add_stats_model.setEnabled(False)

        self.DP4_stat_yn.stateChanged.connect(self.DP4toggle)

        self.DP5_stat_yn.stateChanged.connect(self.DP5toggle)

        self.Add_stats_model.clicked.connect(self.addstats)

        #####################################################################################running PyDP4 in new thread

        # self.Gobutton.clicked.connect(self.Gotoggle)

        self.Gobutton.clicked.connect(self.get_current_values)

        # make worker object

        self.worker = PyDP4WorkerObject()

        # make thread

        self.thread = QtCore.QThread()

        # move the worker object to the ne thread

        self.worker.moveToThread(self.thread)

        # connect workers finished signal to quit thread

        self.worker.finished.connect(self.thread.quit)

        # connect workers finished signal to enable go button

        self.worker.finished.connect(self.Gotoggle)

        self.worker.finished.connect(self.enabletabs)

        # connect start background signal to background job slot

        self.Gobutton.clicked.connect(self.start_PyDP4)

        # start pydp4 give start signal

        # go button press disables go button

        self.Gobutton.clicked.connect(self.Gotoggle)

        # start signal starts pydp4

        self.signal_start_PyDP4.connect(self.worker.runPyDP4)

        self.DFT_settings = DFT_advanced_settings()

        self.DFT_advanced.clicked.connect(self.DFT_pop)

        self.MM_settings = MM_advanced_settings()

        self.MM_advanced.clicked.connect(self.MM_pop)

    def DFT_pop(self):

        self.DFT_settings.show()


    def MM_pop(self):

        self.MM_settings.show()

    def enabletabs(self):

        ui.table_widget.addplottabs()

    def addoutputfolder(self):

        # filename = QtWidgets.QFileDialog.getOpenFileName()
        filename = QtWidgets.QFileDialog.getExistingDirectory()

        # self.NMR_list.addItem(str(filename[0].split("/")[-1]))
        self.Output_list.clear()
        self.Output_list.addItem(filename)
        self.Output_folder = Path(filename)

    def get_current_values(self):

        import PyDP4
        import datetime
        import getpass

        self.settings = PyDP4.Settings()

        self.settings.GUIRunning = True

        now = datetime.datetime.now()
        self.settings.StartTime = now.strftime('%d%b%H%M')

        self.settings.user = getpass.getuser()
        self.settings.DarwinScrDir.replace('/u/', self.settings.user)

        # Read config file and fill in settings in from that
        self.settings = PyDP4.ReadConfig(self.settings)

        # add output folder

        self.settings.OutputFolder = self.Output_folder

        self.log_file = open(self.Output_folder / "DP4_log.log", "w+")

        self.settings.InputFilesPaths = self.Structure_paths

        # copy structures to output folder

        for f in self.Structure_paths:

            if not Path(self.Output_folder / f).exists():
                shutil.copyfile(f, self.settings.OutputFolder / f)

        # add structures

        Smiles = []
        Smarts = []
        InChIs = []

        for index in range(self.Structure_list.count()):

            list_text = self.Structure_list.item(index).text()

            if list_text != '':

                # check if file is text or sdf

                if list_text.endswith('.sdf'):

                    self.settings.InputFiles.append(list_text[:-4])

                # else its a text file - work out if it contains smiles smarts or inchis

                elif (list_text.endswith('Smiles')) or (list_text.endswith('smiles')):

                    Smiles.append(list_text)

                elif (list_text.endswith('Smarts')) or (list_text.endswith('smarts')):

                    Smarts.append(list_text)

                elif (list_text.endswith('InChI')) or (list_text.endswith('InChi')) or (list_text.endswith('inchi')):

                    InChIs.append(list_text)

                else:

                    print(
                        "file type not recognised to use smiles, smarts or inchi input please use .smiles, .smarts or .inchi file extension respectively")

            if len(Smiles) == 1:
                self.settings.Smiles = Smiles[0]

            elif len(Smiles) > 1:

                # if the user has added more than one Smiles string make a new file and concatenate them

                SmilesFile = open(self.settings.OutputFolder / "Smiles_Input.smiles", "w+")

                AllSmiles = []

                for f in Smiles:

                    for line in open(self.settings.OutputFolder / f).readlines():
                        AllSmiles.append(line.strip())

                for s in AllSmiles[:-1]:
                    SmilesFile.write(s + "\n")

                SmilesFile.write(AllSmiles[-1])

                SmilesFile.close()

                self.settings.Smiles = "Smiles_Input.smiles"

            if len(Smarts) == 1:
                self.settings.Smarts = Smarts[0]

            elif len(Smarts) > 1:

                # if the user has added more than one Smiles string make a new file and concatenate them

                SmartsFile = open(self.settings.OutputFolder / "Smarts_Input.smarts", "w+")

                AllSmarts = []

                for f in Smarts:

                    for line in open(self.settings.OutputFolder / f).readlines():
                        AllSmarts.append(line.strip())

                for s in AllSmarts[:-1]:
                    SmartsFile.write(s + "\n")

                SmartsFile.write(AllSmarts[-1])

                SmartsFile.close()

                self.settings.Smarts = "Smarts_Input.smarts"

            if len(InChIs) == 1:
                self.settings.InChIs = InChIs[0]

            elif len(InChIs) > 1:

                # if the user has added more than one Smiles string make a new file and concatenate them

                InchIsFile = open(self.settings.OutputFolder / "InchIs_Input.inchi", "w+")

                AllInchIs = []

                for f in InChIs:

                    for line in open(self.settings.OutputFolder / f).readlines():
                        AllInchIs.append(line.strip())

                for s in AllInchIs[:-1]:
                    InchIsFile.write(s + "\n")

                InchIsFile.write(AllInchIs[-1])

                InchIsFile.close()

                self.settings.InChIs = "InchIs_Input.smarts"


        for f in self.Structure_paths:

            if not Path(self.Output_folder / f).exists():
                shutil.copyfile(f, self.settings.OutputFolder / f)

        # add NMR

        self.settings.NMRsource = self.NMR_paths

        # get workflow information

        # solvent

        if self.Solvent_yn.isChecked() == 1:
            self.settings.Solvent = self.solvent_drop.currentText()

        # generate diastereomers

        self.settings.Workflow = ''

        if self.CleanUp_yn.isChecked() == 1:
            self.settings.Workflow += 'c'

        if self.Gen_diastereomers_yn.isChecked() == 1:
            self.settings.Workflow += 'g'

        # molecular mechanics

        if self.MM_yn.isChecked() == 1:

            self.settings.Workflow += 'm'

            if self.MM_drop.currentText() == "MacroModel":

                self.settings.MM = 'm'

            else:

                self.settings.MM = 't'

        if self.DFT_yn.isChecked() == 1:

            if self.DFT_drop.currentText() == "Gaussian":
                self.settings.DFT = 'g'

            else:
                self.settings.DFT = 'n'

            # Split single point

            if self.DFT_settings.Energy_yn.isChecked() == 1:
                self.settings.Workflow += 'e'
                self.settings.eBasisSet = self.DFT_settings.Energy_basis_drop.currentText()
                self.settings.eFunctional = self.DFT_settings.Energy_functional_drop.currentText()

            if self.DFT_settings.DFTGeom_yn.isChecked() == 1:
                self.settings.Workflow += 'o'
                self.settings.oBasisSet = self.DFT_settings.DFT_geom_basis_drop.currentText()
                self.settings.oFunctional = self.DFT_settings.DFT_geom_functional_drop.currentText()

            if self.DFT_settings.NMR_calc_yn.isChecked() == 1:
                self.settings.Workflow += 'n'
                self.settings.nBasisSet = self.DFT_settings.NMR_basis_drop.currentText()
                self.settings.nFunctional = self.DFT_settings.NMR_functional_drop.currentText()

        if self.DP4_stat_yn.isChecked():

            self.settings.Workflow += 's'

            if self.Stats_list.item(0) != None:
                self.settings.StatsParamFile = self.Stats_list.item(0).text()
                self.settings.StatsModel = 'm'

        if self.DP5_stat_yn.isChecked():

            self.settings.Workflow += 'w'

        elif self.Assignment_yn.isChecked():

            self.settings.Workflow += 'a'

        self.settings.ScriptDir = os.path.dirname(os.path.realpath(sys.argv[0]))

    def start_PyDP4(self):

        # No harm in calling thread.start() after the thread is already started.

        # start the thread

        self.thread.start()

        self.thread.finished.connect(self.Gotoggle)

        # send background job start signal

        self.signal_start_PyDP4.emit()

        self.thread.disconnect()

    def addstructure(self):

        f = QtWidgets.QFileDialog.getOpenFileName()[0]

        if f:
            filename = Path(f)

            self.Structure_list.addItem(filename.name)
            self.Structure_paths.append(filename)

    def removestructure(self):

        item = self.Structure_list.selectedItems()
        if not item:
            return
        for i in item:
            self.Structure_list.takeItem(self.Structure_list.row(i))
            self.Structure_paths.pop(self.Structure_list.row(i))

    def addNMR(self):

        # filename = QtWidgets.QFileDialog.getOpenFileName()

        i = QtWidgets.QFileDialog.getExistingDirectory()

        if i:

            filename = Path(i)

            p_switch = 0

            c_switch = 0

            for f in filename.iterdir():

                if f.name == "Carbon" or f.name == "carbon" or f.name == "Carbon.dx" or f.name == "carbon.dx":
                    self.NMR_list.addItem(f.name)
                    self.NMR_paths.append(f)
                    c_switch = 1

                elif f.name == "Proton" or f.name == "proton" or f.name == "Proton.dx" or f.name == "proton.dx":
                    self.NMR_list.addItem(f.name)
                    self.NMR_paths.append(f)
                    p_switch = 1

                if p_switch == 1 and c_switch == 1:
                    break

            # self.NMR_list.addItem(str(filename[0].split("/")[-1]))

            if p_switch == 0 and c_switch == 0:
                self.NMR_list.addItem(filename.name)
                self.NMR_paths.append(filename.name)

            self.Add_NMR_desc.setEnabled(False)

    def addNMRdesc(self):

        f = QtWidgets.QFileDialog.getOpenFileName()[0]

        if f:
            filename = Path(f)

            self.NMR_list.addItem(filename.name)
            self.NMR_paths.append(filename)

            self.Add_NMR.setEnabled(False)
            self.Add_NMR_desc.setEnabled(False)


    def remove_all_NMR(self):

        self.NMR_list.clear()
        self.NMR_paths = []
        self.Add_NMR.setEnabled(True)
        self.Add_NMR_desc.setEnabled(True)

    def removeNMR(self):

        item = self.NMR_list.selectedItems()
        if not item:
            return
        for i in item:
            self.NMR_list.takeItem(self.NMR_list.row(i))
            self.NMR_paths.pop(self.NMR_list.row(i))

        if len(self.NMR_paths) == 0:
            self.Add_NMR.setEnabled(True)
            self.Add_NMR_desc.setEnabled(True)


    def addstats(self):
        filename = QtWidgets.QFileDialog.getOpenFileName()
        self.Stats_list.addItem(str(filename[0]))

    def solventtoggle(self, state):

        if state > 0:
            self.solvent_drop.setEnabled(True)
            #self.Gen_diastereomers_yn.setChecked(True)

        else:
            self.solvent_drop.setEnabled(False)
            self.MM_yn.setChecked(False)

    def MMtoggle(self, state):

        if state > 0:

            #self.Gen_diastereomers_yn.setChecked(True)

            self.Solvent_yn.setChecked(True)
            self.solvent_drop.setEnabled(True)
            self.MM_drop.setEnabled(True)
            self.MM_advanced.setEnabled(True)


        else:
            self.solvent_drop.setEnabled(False)
            self.Solvent_yn.setChecked(False)
            self.MM_drop.setEnabled(False)
            self.MM_advanced.setEnabled(False)
            self.DFT_yn.setChecked(False)

    def DFTtoggle(self,state):

        if state > 0:

            # self.Gen_diastereomers_yn.setChecked(True)

            self.Solvent_yn.setChecked(True)
            self.solvent_drop.setEnabled(True)
            self.MM_yn.setChecked(True)
            self.MM_drop.setEnabled(True)
            self.DFT_drop.setEnabled(True)
            self.DFT_advanced.setEnabled(True)
            self.DFT_settings.NMR_calc_yn.setChecked(True)

        else:
            self.DFT_advanced.setEnabled(False)
            self.DFT_drop.setEnabled(False)
            self.solvent_drop.setEnabled(False)
            self.Solvent_yn.setChecked(False)
            self.MM_yn.setChecked(False)
            self.MM_drop.setEnabled(False)
            self.Assignment_yn.setChecked(False)
            self.DP4_stat_yn.setChecked(False)

            self.DFT_settings.DFTGeom_yn.setChecked(False)
            self.DFT_settings.Energy_yn.setChecked(False)
            self.DFT_settings.NMR_calc_yn.setChecked(False)

    def Assignment_toggle(self, state):

        if state > 0:

            self.DFT_yn.setChecked(True)
            self.Solvent_yn.setChecked(True)
            self.solvent_drop.setEnabled(True)
            #self.Gen_diastereomers_yn.setChecked(True)
            self.MM_yn.setChecked(True)

        else:
            self.DP4_stat_yn.setChecked(False)
            self.Add_stats_model.setEnabled(False)

            self.DFT_yn.setChecked(False)
            self.Solvent_yn.setChecked(False)
            self.solvent_drop.setEnabled(False)
            #self.Gen_diastereomers_yn.setChecked(True)
            self.MM_yn.setChecked(False)

    def DP4toggle(self, state):

        if state > 0:
            self.Add_stats_model.setEnabled(True)

            self.DFT_yn.setChecked(True)
            self.Solvent_yn.setChecked(True)
            self.solvent_drop.setEnabled(True)
            #self.Gen_diastereomers_yn.setChecked(True)
            self.MM_yn.setChecked(True)
            self.Assignment_yn.setChecked(True)

        else:
            self.Add_stats_model.setEnabled(False)
            self.DFT_yn.setChecked(False)
            self.Solvent_yn.setChecked(False)
            self.solvent_drop.setEnabled(False)
            #self.Gen_diastereomers_yn.setChecked(True)
            self.MM_yn.setChecked(False)
            self.Assignment_yn.setChecked(False)

    def DP5toggle(self, state):

        if state > 0:
            self.DFT_yn.setChecked(True)
            self.Solvent_yn.setChecked(True)
            self.solvent_drop.setEnabled(True)
            #self.Gen_diastereomers_yn.setChecked(True)
            self.MM_yn.setChecked(True)
            self.Assignment_yn.setChecked(True)

        else:
            self.DFT_yn.setChecked(False)
            self.Solvent_yn.setChecked(False)
            self.solvent_drop.setEnabled(False)
            #self.Gen_diastereomers_yn.setChecked(True)
            self.MM_yn.setChecked(False)
            self.Assignment_yn.setChecked(False)

    def append_text(self, text):
        self.Output_box.moveCursor(QtGui.QTextCursor.End)
        self.Output_box.insertPlainText(text)
        # self.log_file.write(text)

        self.log_file.write(text)

    def Gotoggle(self):

        if self.Gobutton.isEnabled() == True:

            self.Gobutton.setEnabled(False)

        else:

            self.Gobutton.setEnabled(True)


class DFT_advanced_settings(QtWidgets.QWidget):

    def __init__(self):
        super(DFT_advanced_settings, self).__init__()
        self.layout = QtWidgets.QVBoxLayout(self)

        self.title = QtWidgets.QLabel(self)
        self.title.setText("DFT Advanced Options")
        self.layout.addWidget(self.title)

        ##################################### DFT Geom Optimisation

        self.DFTGeom_yn = QtWidgets.QCheckBox(self)
        self.DFTGeom_yn.setText("DFT Geometry Optimisation")
        self.layout.addWidget(self.DFTGeom_yn)

        self.DFT_geom_functional_drop_label = QtWidgets.QLabel(self)
        self.DFT_geom_functional_drop_label.setText("Functional")
        self.layout.addWidget(self.DFT_geom_functional_drop_label)

        self.DFT_geom_functional_drop = QtWidgets.QComboBox(self)
        DFTopt_functional = ['B3LYP', 'm062x', 'mPW1PW91']
        self.DFT_geom_functional_drop.addItems(DFTopt_functional)
        self.layout.addWidget(self.DFT_geom_functional_drop)

        self.DFT_geom_basis_drop_label = QtWidgets.QLabel(self)
        self.DFT_geom_basis_drop_label.setText("Basis Set")
        self.layout.addWidget(self.DFT_geom_basis_drop_label)

        self.DFT_geom_basis_drop = QtWidgets.QComboBox(self)
        DFTopt_basis = ['6-31g(d)', '6-311g(d)', 'def2svp', 'def2tzvp']
        self.DFT_geom_basis_drop.addItems(DFTopt_basis)
        self.layout.addWidget(self.DFT_geom_basis_drop)

        ##################################### DFT Energy

        self.Energy_yn = QtWidgets.QCheckBox(self)
        self.Energy_yn.setText("Split Single Point DFT")
        self.layout.addWidget(self.Energy_yn)

        self.Energy_functional_drop_label = QtWidgets.QLabel(self)
        self.Energy_functional_drop_label.setText("Functional")
        self.layout.addWidget(self.Energy_functional_drop_label)

        self.Energy_functional_drop = QtWidgets.QComboBox(self)
        Energy_functional = ['m062x', 'mPW1PW91', 'B3LYP']
        self.Energy_functional_drop.addItems(Energy_functional)
        self.layout.addWidget(self.Energy_functional_drop)

        self.Energy_basis_drop_label = QtWidgets.QLabel(self)
        self.Energy_basis_drop_label.setText("Basis Set")
        self.layout.addWidget(self.Energy_basis_drop_label)

        self.Energy_basis_drop = QtWidgets.QComboBox(self)
        Energy_basis = ['6-31g(d)', '6-311g(d)', 'def2svp', 'def2tzvp']
        self.Energy_basis_drop.addItems(Energy_basis)
        self.layout.addWidget(self.Energy_basis_drop)

        ##################################### DFT NMR

        self.NMR_calc_yn = QtWidgets.QCheckBox(self)
        self.NMR_calc_yn.setText("NMR Calculations")
        self.layout.addWidget(self.NMR_calc_yn)

        self.NMR_functional_drop_label = QtWidgets.QLabel(self)
        self.NMR_functional_drop_label.setText("Functional")
        self.layout.addWidget(self.NMR_functional_drop_label)

        self.NMR_functional_drop = QtWidgets.QComboBox(self)
        NMR_functional = ['mPW1PW91', 'B3LYP', 'M062X']
        self.NMR_functional_drop.addItems(NMR_functional)
        self.layout.addWidget(self.NMR_functional_drop)

        self.NMR_basis_drop_label = QtWidgets.QLabel(self)
        self.NMR_basis_drop_label.setText("Basis Set")
        self.layout.addWidget(self.NMR_basis_drop_label)

        self.NMR_basis_drop = QtWidgets.QComboBox(self)
        NMR_basis = ['6-311g(d)', '6-31g(d)', 'def2svp', 'def2tzvp']
        self.NMR_basis_drop.addItems(NMR_basis)
        self.layout.addWidget(self.NMR_basis_drop)

        ###################################### methods

        # selecting DFT geometry opt box

        self.DFT_geom_functional_drop.setEnabled(False)

        self.DFT_geom_basis_drop.setEnabled(False)

        self.DFTGeom_yn.stateChanged.connect(self.DFTopttoggle)

        # selecting Split Single point box

        self.Energy_functional_drop.setEnabled(False)

        self.Energy_basis_drop.setEnabled(False)

        self.Energy_yn.stateChanged.connect(self.Energytoggle)

        # selecting NMR box

        self.NMR_functional_drop.setEnabled(False)

        self.NMR_basis_drop.setEnabled(False)

        self.NMR_calc_yn.stateChanged.connect(self.NMRtoggle)


    def Energytoggle(self, state):

        if state > 0:
            self.Energy_functional_drop.setEnabled(True)
            self.Energy_basis_drop.setEnabled(True)

        else:
            self.Energy_functional_drop.setEnabled(False)
            self.Energy_basis_drop.setEnabled(False)

    def DFTopttoggle(self, state):

        if state > 0:
            self.DFT_geom_functional_drop.setEnabled(True)
            self.DFT_geom_basis_drop.setEnabled(True)

        else:
            self.DFT_geom_functional_drop.setEnabled(False)
            self.DFT_geom_basis_drop.setEnabled(False)

    def NMRtoggle(self, state):

        if state > 0:
            self.NMR_functional_drop.setEnabled(True)
            self.NMR_basis_drop.setEnabled(True)

        else:
            self.NMR_functional_drop.setEnabled(False)
            self.NMR_basis_drop.setEnabled(False)


class MM_advanced_settings(QtWidgets.QWidget):

    def __init__(self):
        super(MM_advanced_settings, self).__init__()

        self.layout = QtWidgets.QVBoxLayout(self)

        self.title = QtWidgets.QLabel(self)
        self.title.setText("MM Advanced Options")
        self.layout.addWidget(self.title)


class ProtonPlotTab(QtWidgets.QWidget):

    def __init__(self):

        super(ProtonPlotTab, self).__init__()

        self.cwd = Path(os.getcwd())

        # create a figure instance

        self.figure = Figure()

        # create the canvas widget thats displays figure

        self.canvas = FigureCanvas(self.figure)

        # make the navigation widget - this takes the canvas widget as the parent

        self.toolbar = NavigationToolbar(self.canvas, self)

        self.IsomerSelect = QtWidgets.QComboBox(self)

        self.IsomerSelect.addItems(self.Isomer_number())

        self.IsomerSelect.currentIndexChanged.connect(self.PlotProton)
        self.image = QSvgWidget()

        self.widget1 = QtWidgets.QWidget(self)

        self.widget2 = QtWidgets.QWidget(self)

        self.hl1 = QtWidgets.QVBoxLayout(self.widget1)

        self.hl2 = QtWidgets.QVBoxLayout(self.widget2)

        self.hl1.addWidget(self.toolbar)

        self.hl1.addWidget(self.canvas)

        self.hl2.addWidget(self.IsomerSelect)

        self.hl2.addWidget(self.image)

        self.widget1.setGeometry(QtCore.QRect(300, 0, 1200, 875))

        # self.IsomerSelect.setGeometry(QtCore.QRect(0,0,100,50))

        # self.image.setGeometry(QtCore.QRect(0,50,300,300))

        self.widget2.setGeometry(QtCore.QRect(0, 0, 300, 400))

        self.canvas.mpl_connect('button_press_event', self.selectpoint)

        ################################################################################################################

        if ui.table_widget.Tab1.settings.OutputFolder == '':

            pdir = self.cwd / "Pickles"

        else:
            pdir = ui.table_widget.Tab1.settings.OutputFolder / "Pickles"

        if os.path.isfile(pdir / ui.table_widget.Tab1.settings.InputFiles[0] / "protondata"):
            self.protondata = pickle.load(
                Path(pdir / ui.table_widget.Tab1.settings.InputFiles[0] / "protondata").open(mode="rb"))

        self.xdata = self.protondata["xdata"]

        self.ydata = self.protondata["ydata"]

        self.centres = self.protondata["centres"]

        self.exp_peaks = self.protondata["exppeaks"]

        self.peak_regions = self.protondata["peakregions"]

        self.cummulative_vectors = self.protondata["cummulativevectors"]

        self.integral_sum = self.protondata["integralsum"]

        self.integrals = self.protondata["integrals"]

        self.sim_regions = self.protondata["sim_regions"]

        self.PlotProton()

        ################################################################################################################

    def Isomer_number(self):

        Isomer_list = []

        for c, i in enumerate(ui.table_widget.Tab1.worker.Isomers):
            Isomer_list.append("Isomer " + str(c + 1))

        return Isomer_list

    def PlotProton(self):

        self.RenderImage([], None)

        self.isomerindex = int(str(self.IsomerSelect.currentText())[-1]) -1

        self.isomer = ui.table_widget.Tab1.worker.Isomers[self.isomerindex]

        self.assigned_shifts = self.isomer.Hshifts

        self.assigned_peaks = []

        for peak in self.isomer.Hexp:

            if peak != '':
                self.assigned_peaks.append(peak)

        self.assigned_labels = self.isomer.Hlabels

        # check if pickle and DP4output files are in place

        if ui.table_widget.Tab1.worker.settings.OutputFolder == '':

            pdir = Path.cwd() / "Pickles"

        else:

            pdir = ui.table_widget.Tab1.worker.settings.OutputFolder / "Pickles"

        if Path(pdir / ui.table_widget.Tab1.worker.settings.InputFiles[0] / "protondata").exists():

            self.figure.clear()

            fig = self.figure.add_subplot(111)

            self.figure.subplots_adjust(left=0.05, right=0.95, bottom=0.07, top=0.95, wspace=0.05, hspace=0.05)

            #################################### will probs need to fix sorting here

            fig.set_xlim([10, 0])

            fig.set_xlabel("ppm")

            fig.plot(self.xdata, self.ydata, label='data', color='grey')

            set_exp = sorted(list(set(self.exp_peaks)))[::-1]

            # simulate_spectrum(xdata, assigned_shifts, assigned_peaks, set_exp)

            ##############################

            for ind, shift in enumerate(self.assigned_shifts):
                exp_p = self.assigned_peaks[ind]

                ind2 = set_exp.index(exp_p)
                y = lorentzian(self.xdata, 0.001, shift, 0.2)

                fig.plot(self.xdata, y + 1.05, color='C' + str(ind2 % 10))

            ##############################

            fig.axhline(1.05, color='grey')

            # plt integral information

            prev = 15

            count = 0

            for index in range(0, len(self.peak_regions)):

                if abs(prev - self.xdata[self.centres[index]]) < 0.45:
                    count += 1
                else:
                    count = 0
                    prev = self.xdata[self.centres[index]]

                fig.annotate(str(self.integrals[index]) + ' Hs',
                             xy=(self.xdata[self.centres[index]], -(0.1) - 0.1 * count), color='C' + str(index % 10))

                fig.annotate(str(round(self.xdata[self.centres[index]], 3)) + ' ppm',
                             xy=(self.xdata[self.centres[index]], -(0.15) - 0.1 * count), color='C' + str(index % 10))

                fig.plot(self.xdata[self.peak_regions[index]],
                         self.cummulative_vectors[index] + self.integral_sum[index],
                         color='C' + str(index % 10),
                         linewidth=2)

            for index in range(0, len(self.peak_regions) - 1):
                fig.plot([self.xdata[self.peak_regions[index][-1]], self.xdata[self.peak_regions[index + 1][0]]],
                         [self.integral_sum[index + 1], self.integral_sum[index + 1]], color='grey')

            for index, region in enumerate(self.peak_regions):
                fig.plot(self.xdata[region], self.sim_regions[index], color='C' + str(index % 10))

            # plt.legend()

            ### plotting assignment

            fig.set_yticks([], [])

            # fig.set_title(str(ui.table_widget.Tab1.settings.InputFiles[0]) +
            #             "\nProton NMR of Isomer " + str(isomerindex + 1) + "\n Number of Peaks Found = " + str(
            #  len(exp_peaks)))

            # plot assignments

            for ind1, peak in enumerate(self.assigned_peaks):
                fig.plot([peak, self.assigned_shifts[ind1]],
                         [1, 1.05], linewidth=0.5, color='cyan')

            # annotate peak locations

            for x, txt in enumerate(self.exp_peaks):

                if self.exp_peaks[x] in self.assigned_peaks:

                    color = 'C1'

                else:

                    color = 'grey'

                fig.plot(txt, -0.02, 'o', color=color)

            # annotate shift positions

            prev = 0

            count = 0

            s = np.argsort(np.array(self.assigned_shifts))

            s_assigned_shifts = np.array(self.assigned_shifts)[s]

            s_assigned_labels = np.array(self.assigned_labels)[s]

            s_assigned_peaks = np.array(self.assigned_peaks)[s]

            for x, txt in enumerate(s_assigned_labels[::-1]):

                w = np.where(set_exp == s_assigned_peaks[::-1][x])[0][0]

                color = w % 10

                if abs(prev - s_assigned_shifts[::-1][x]) < 0.2:
                    count += 1

                else:
                    count = 0
                    prev = s_assigned_shifts[::-1][x]

                fig.annotate(txt, (s_assigned_shifts[::-1][x], + 1.25 + 0.05 * count), color='C' + str(color))

            # ax1.plot(picked_peaks_ppm,ydata[picked_peaks],
            #        'co', label='Picked Peaks')

            fig.set_ylim([-0.5, 2.0])

            self.canvas.draw()

        else:
            pass

    def selectpoint(self, event):

        if event.dblclick:

            self.xpos = event.xdata

            self.ypos = event.ydata

            # find the cloest point to the click

            if self.xpos is None and self.xpos is None:

                self.PlotProton()

            elif self.ypos < 1:

                mindis = np.argmin(abs(self.xdata[self.centres] - self.xpos))

                self.PlotProtonSelected(mindis)

                # find which protons this peak has been assigned to

                p = np.where([round(i, 4) for i in self.assigned_peaks] == round(self.xdata[self.centres[mindis]], 4))[
                    0]

                la = [int(self.assigned_labels[j][1:]) - 1 for j in p]

                self.RenderImage(la, mindis)

            else:

                # find closest assigned shift

                m = np.argmin(abs(np.array(self.assigned_shifts) - self.xpos))

                # find which peak this is assigned to

                p = self.assigned_peaks[m]

                mindis = np.argmin(abs(p - self.xdata[self.centres]))

                p = np.where(self.assigned_peaks == round(self.xdata[self.centres[mindis]], 4))[0]

                la = [int(self.assigned_labels[j][1:]) - 1 for j in p]

                self.PlotProtonSelected(mindis)

                self.RenderImage(la, mindis)

    def PlotProtonSelected(self, mindis):

        # check if pickle and DP4output files are in place

        if ui.table_widget.Tab1.worker.settings.OutputFolder == '':

            pdir = Path.cwd() / "Pickles"

        else:

            pdir = ui.table_widget.Tab1.worker.settings.OutputFolder / "Pickles"

        if Path(pdir / ui.table_widget.Tab1.worker.settings.InputFiles[0] / "protondata").exists():

            self.figure.clear()

            fig = self.figure.add_subplot(111)

            self.figure.subplots_adjust(left=0.05, right=0.95, bottom=0.07, top=0.95, wspace=0.05, hspace=0.05)

            assigned_shifts = self.isomer.Hshifts

            assigned_peaks = []

            for peak in self.isomer.Hexp:

                if peak != '':
                    assigned_peaks.append(peak)

            assigned_labels = self.isomer.Hlabels

            #################################### will probs need to fix sorting here

            fig.set_xlim([10, 0])

            fig.set_xlabel("ppm")

            fig.plot(self.xdata, self.ydata, label='data', color='grey', alpha=0.5)

            set_exp = sorted(list(set(self.exp_peaks)))[::-1]

            # simulate_spectrum(xdata, assigned_shifts, assigned_peaks, set_exp)

            ##############################

            for ind, shift in enumerate(assigned_shifts):

                exp_p = assigned_peaks[ind]

                ind2 = set_exp.index(exp_p)

                y = lorentzian(self.xdata, 0.001, shift, 0.2)

                if ind2 == mindis:

                    fig.plot(self.xdata, y + 1.05, color='C' + str(ind2 % 10))

                else:

                    fig.plot(self.xdata, y + 1.05, color='grey', alpha=0.5)

            ##############################

            fig.axhline(1.05, color='grey')

            # plt integral information

            prev = 15

            count = 0

            for index, region in enumerate(self.peak_regions):
                fig.plot(self.xdata[region], self.sim_regions[index], color='grey')

            fig.plot(self.xdata[self.peak_regions[mindis]], self.sim_regions[mindis], color='C' + str(mindis))

            fig.annotate(str(self.integrals[mindis]) + ' Hs',
                         xy=(self.xdata[self.centres[mindis]], -(0.1) - 0.1 * count), color='C' + str(mindis))

            fig.annotate(str(round(self.xdata[self.centres[mindis]], 3)) + ' ppm',
                         xy=(self.xdata[self.centres[mindis]], -(0.15) - 0.1 * count), color='C' + str(mindis))

            fig.plot(self.xdata[self.peak_regions[mindis]],
                     self.cummulative_vectors[mindis] + self.integral_sum[mindis],
                     color='C' + str(mindis % 10),
                     linewidth=2)
            # plt.legend()

            ### plotting assignment

            fig.set_yticks([], [])

            # fig.set_title(str(ui.table_widget.Tab1.settings.InputFiles[0]) +
            #             "\nProton NMR of Isomer " + str(isomerindex + 1) + "\n Number of Peaks Found = " + str(
            #  len(exp_peaks)))

            # plot assignments

            # annotate shift positions

            prev = 0

            count = 0

            s = np.argsort(np.array(assigned_shifts))

            s_assigned_shifts = np.array(assigned_shifts)[s]

            s_assigned_labels = np.array(assigned_labels)[s]

            s_assigned_peaks = np.array(assigned_peaks)[s]

            for x, txt in enumerate(s_assigned_labels[::-1]):

                w = np.where(set_exp == s_assigned_peaks[::-1][x])[0][0]

                if w == mindis:

                    color = w % 10

                    if abs(prev - s_assigned_shifts[::-1][x]) < 0.2:
                        count += 1

                    else:
                        count = 0
                        prev = s_assigned_shifts[::-1][x]

                    fig.annotate(txt, (s_assigned_shifts[::-1][x], + 1.25 + 0.05 * count), color='C' + str(color))

            # ax1.plot(picked_peaks_ppm,ydata[picked_peaks],
            #        'co', label='Picked Peaks')

            fig.set_ylim([-0.5, 2.0])

            for x, txt in enumerate(self.exp_peaks):

                if self.exp_peaks[x] in self.assigned_peaks:

                    color = 'C1'

                else:

                    color = 'grey'

                fig.plot(txt, -0.02, 'o', color=color)

            self.canvas.draw()

        else:
            pass

    def RenderImage(self, atom, color):

        colors = [(0.12, 0.47, 0.71), (1.0, 0.5, 0.05), (0.17, 0.63, 0.17), (0.84, 0.15, 0.16), (0.58, 0.4, 0.74),
                  (0.55, 0.34, 0.29), (0.89, 0.47, 0.76), (0.5, 0.5, 0.5), (0.74, 0.74, 0.13), (0.09, 0.75, 0.81)]

        highlight = {}

        for i in atom:
            highlight[i] = colors[color % 10]

        isomerindex = int(str(self.IsomerSelect.currentText())[-1]) -1

        m = Chem.MolFromMolFile(str(ui.table_widget.Tab1.worker.settings.InputFilesPaths[isomerindex]).split('.sdf')[0] + '.sdf',removeHs=False)

        Chem.Compute2DCoords(m)

        drawer = rdMolDraw2D.MolDraw2DSVG(300, 300)

        drawer.DrawMolecule(m, highlightAtoms=atom, highlightAtomColors=highlight)

        drawer.FinishDrawing()

        svg = drawer.GetDrawingText().replace('svg:', '')

        svg_bytes = bytearray(svg, encoding='utf-8')

        self.image.renderer().load(svg_bytes)

        self.image.setGeometry(QtCore.QRect(0, 50, 300, 300))

        ui.update()

        # f = open("f.svg", "w+")

        # f.write(str(svg))


class CarbonPlotTab(QtWidgets.QWidget):

    def __init__(self):

        super(CarbonPlotTab, self).__init__()

        # create a figure instance

        self.figure = Figure()

        # create the canvas widget thats displays figure

        self.canvas = FigureCanvas(self.figure)

        # make the navigation widget - this takes the canvas widget as the parent

        self.toolbar = NavigationToolbar(self.canvas, self)

        self.IsomerSelect = QtWidgets.QComboBox(self)

        self.IsomerSelect.addItems(self.Isomer_number())

        self.IsomerSelect.currentIndexChanged.connect(self.PlotCarbon)

        self.image = QSvgWidget()

        self.widget1 = QtWidgets.QWidget(self)

        self.widget2 = QtWidgets.QWidget(self)

        self.hl1 = QtWidgets.QVBoxLayout(self.widget1)

        self.hl2 = QtWidgets.QVBoxLayout(self.widget2)

        self.hl1.addWidget(self.toolbar)

        self.hl1.addWidget(self.canvas)

        self.hl2.addWidget(self.IsomerSelect)

        self.hl2.addWidget(self.image)

        self.widget1.setGeometry(QtCore.QRect(300, 0, 1200, 875))

        self.widget2.setGeometry(QtCore.QRect(0, 0, 300, 400))

        self.canvas.mpl_connect('button_press_event', self.selectpoint)

        #############################

        if ui.table_widget.Tab1.settings.OutputFolder == '':

            pdir = Path.cwd() / "Pickles"

        else:
            pdir = ui.table_widget.Tab1.settings.OutputFolder / "Pickles"

        if Path(pdir / ui.table_widget.Tab1.settings.InputFiles[0] / "carbondata").exists():
            self.carbondata = pickle.load(
                Path(pdir / ui.table_widget.Tab1.settings.InputFiles[0] / "carbondata").open(mode="rb"))

        self.xdata = self.carbondata["xdata"]

        self.ydata = self.carbondata["ydata"]

        self.exppeaks = self.carbondata["exppeaks"]

        self.simulated_ydata = self.carbondata["simulated_ydata"]

        self.removed = self.carbondata["removed"]

        self.PlotCarbon()

        #############################

    def Isomer_number(self):

        Isomer_list = []

        for c, i in enumerate(ui.table_widget.Tab1.worker.Isomers):
            Isomer_list.append("Isomer " + str(c + 1))

        return Isomer_list

    def PlotCarbon(self):

        self.RenderImage([])

        if ui.table_widget.Tab1.worker.settings.OutputFolder == '':

            pdir = Path.cwd() / "Pickles"

        else:

            pdir = ui.table_widget.Tab1.worker.settings.OutputFolder / "Pickles"

        if Path(pdir / ui.table_widget.Tab1.worker.settings.InputFiles[0] / "carbondata").exists():

            self.figure.clear()

            fig = self.figure.add_subplot(111)

            self.figure.subplots_adjust(left=0.05, right=0.95, bottom=0.07, top=0.95, wspace=0.05, hspace=0.05)

            self.isomerindex = int(str(self.IsomerSelect.currentText())[-1]) -1

            self.isomer = ui.table_widget.Tab1.worker.Isomers[self.isomerindex]
            self.assigned_shifts = self.isomer.Cshifts

            self.assigned_peaks = []

            for peak in self.isomer.Cexp:

                if peak != '':
                    self.assigned_peaks.append(peak)

            self.assigned_labels = self.isomer.Clabels

            ###################will probs need to fix sorting here

            exppeaks_ppm = self.xdata[self.exppeaks].tolist()

            shiftlist = self.assigned_shifts

            totallist = exppeaks_ppm + shiftlist

            fig.set_xlim([max(totallist) + 10, min(totallist) - 10])

            fig.plot(self.xdata, self.ydata, color='grey', linewidth=0.75, label='experimental spectrum')
            fig.plot(self.xdata, self.simulated_ydata, label='simulated spectrum')

            fig.set_xlabel('PPM')  # axis labels

            # plt.yticks([], [])
            # fig.set_title(str(ui.table_widget.Tab1.worker.settings.InputFiles[0]) +
            #             "\nCarbon NMR of Isomer " + str(self.isomerindex + 1) + "\n Number of Peaks Found = " + str(
            #  len(self.exppeaks)))

            # plot assignments

            for ind1, peak in enumerate(self.assigned_peaks):
                wh = np.argmin(abs(self.xdata - peak))

                fig.plot([peak, self.assigned_shifts[ind1]],
                         [self.ydata[wh], 1.1], linewidth=0.5, color='cyan')

            prev = round(exppeaks_ppm[0], 2)

            count = 0

            # annotate peak locations

            for x, txt in enumerate([round(i, 2) for i in exppeaks_ppm]):

                if abs(prev - txt) < 5:

                    count += 1
                else:
                    count = 0
                    prev = txt

                if exppeaks_ppm[x] in self.assigned_peaks:
                    color = 'C1'
                else:
                    color = 'grey'

                fig.annotate(txt, (exppeaks_ppm[x], -0.06 - 0.025 * count), color=color)

                fig.plot(exppeaks_ppm[x], self.ydata[self.exppeaks[x]], 'o', color=color)

            if len(self.removed) > 0:
                fig.plot(self.xdata[self.removed],
                         self.simulated_ydata[self.removed], "ro")

            # annotate shift positions

            count = 0

            ####some quick sorting

            argss = np.argsort(self.assigned_shifts)
            sortshifts = np.sort(self.assigned_shifts)[::-1]
            slabels = np.array(self.assigned_labels)[argss][::-1]

            prev = sortshifts[0]

            for x, txt in enumerate(slabels):

                if abs(prev - sortshifts[x]) < 4:
                    count += 1
                else:
                    count = 0
                    prev = sortshifts[x]

                fig.annotate(txt, (sortshifts[x], + 2.05 + 0.05 * count))

            ##########

            simulated_calc_ydata = np.zeros(len(self.xdata))

            for peak in self.assigned_shifts:
                y = np.exp(-0.5 * ((self.xdata - peak) / 0.002) ** 2)
                simulated_calc_ydata += y

            scaling_factor = np.amax(self.simulated_ydata) / np.amax(simulated_calc_ydata)

            simulated_calc_ydata = simulated_calc_ydata * scaling_factor

            #########

            fig.plot(self.xdata, simulated_calc_ydata + 1.1, label='calculated spectrum')

            self.canvas.draw()

    def selectpoint(self, event):

        if event.dblclick:

            self.xpos = event.xdata

            self.ypos = event.ydata

            # find the cloest point to the click

            if self.xpos is None and self.xpos is None:

                self.PlotCarbon()

            elif self.ypos < 1:

                mindis = np.argmin(abs(self.xdata[self.exppeaks] - self.xpos))

                self.PlotCarbonSelected(mindis)

                # find which protons this peak has been assigned to

                p = np.where(self.assigned_peaks == self.xdata[self.exppeaks[mindis]])[0]

                la = [int(self.assigned_labels[j][1:]) - 1 for j in p]

                self.RenderImage(la)

            else:

                # find closest assigned shift

                m = np.argmin(abs(np.array(self.assigned_shifts) - self.xpos))

                # find which peak this is assigned to

                p = self.assigned_peaks[m]

                mindis = np.argmin(abs(p - self.xdata[self.exppeaks]))

                p = np.where(self.assigned_peaks == self.xdata[self.exppeaks[mindis]])[0]

                la = [int(self.assigned_labels[j][1:]) - 1 for j in p]

                self.PlotCarbonSelected(mindis)

                self.RenderImage(la)

    def PlotCarbonSelected(self, mindis):

        if ui.table_widget.Tab1.worker.settings.OutputFolder == '':

            pdir = Path.cwd() / "Pickles"

        else:

            pdir = ui.table_widget.Tab1.worker.settings.OutputFolder / "Pickles"

        if Path(pdir / ui.table_widget.Tab1.worker.settings.InputFiles[0] / "carbondata").exists():

            self.figure.clear()

            fig = self.figure.add_subplot(111)

            self.figure.subplots_adjust(left=0.05, right=0.95, bottom=0.07, top=0.95, wspace=0.05, hspace=0.05)

            exppeaks_ppm = self.xdata[self.exppeaks].tolist()

            shiftlist = self.assigned_shifts

            totallist = exppeaks_ppm + shiftlist

            fig.set_xlim([max(totallist) + 10, min(totallist) - 10])

            fig.plot(self.xdata, self.ydata, color='grey', linewidth=0.75, label='experimental spectrum')

            fig.plot(self.xdata, self.simulated_ydata, label='simulated spectrum')

            fig.set_xlabel('PPM')  # axis labels

            assigned_peak = exppeaks_ppm[mindis]

            s = np.where(np.round(self.assigned_peaks, 8) == np.round(assigned_peak, 8))[0]

            assigned_shift = np.array(self.assigned_shifts)[s]

            assigned_label = np.array(self.assigned_labels)[s]

            for i in assigned_shift:
                fig.plot([assigned_peak, i], [self.ydata[self.exppeaks[mindis]], 1.1], color='cyan')

            # plot assignments

            # for ind1, peak in enumerate(self.assigned_peaks):
            #   wh = np.argmin(abs(self.xdata - peak))

            #  fig.plot([peak, self.assigned_shifts[ind1]],
            #          [self.ydata[wh], 1.1], linewidth=0.5, color='cyan')

            prev = round(exppeaks_ppm[0], 2)

            count = 0

            # annotate peak locations

            for x, txt in enumerate([round(i, 2) for i in exppeaks_ppm]):

                if exppeaks_ppm[x] == assigned_peak:

                    color = 'cyan'

                elif exppeaks_ppm[x] in self.assigned_peaks:

                    color = 'C1'

                else:
                    color = 'grey'

                fig.plot(exppeaks_ppm[x], self.ydata[self.exppeaks[x]], 'o', color=color)

            fig.annotate(str(round(assigned_peak, 2)), (assigned_peak, -0.06), color='cyan')

            if len(self.removed) > 0:
                fig.plot(self.xdata[self.removed],
                         self.simulated_ydata[self.removed], "ro")

            for l, s in zip(assigned_label, assigned_shift):
                fig.annotate(l, (s, 2.05), color="cyan")

            ##########

            simulated_calc_ydata = np.zeros(len(self.xdata))

            for peak in self.assigned_shifts:
                y = np.exp(-0.5 * ((self.xdata - peak) / 0.002) ** 2)

                simulated_calc_ydata += y

            scaling_factor = np.amax(self.simulated_ydata) / np.amax(simulated_calc_ydata)

            for peak in self.assigned_shifts:
                y = np.exp(-0.5 * ((self.xdata - peak) / 0.002) ** 2)

                fig.plot(self.xdata, y * scaling_factor + 1.1, label='calculated spectrum', color='grey', alpha=0.5)

            for peak in assigned_shift:
                y = np.exp(-0.5 * ((self.xdata - peak) / 0.002) ** 2)

                fig.plot(self.xdata, y * scaling_factor + 1.1, label='calculated spectrum', color='C1')

            #########

            self.canvas.draw()

    def RenderImage(self, atom):

        highlight = {}

        for i in atom:
            highlight[i] = (0, 1, 1)

        isomerindex = int(str(self.IsomerSelect.currentText())[-1]) -1

        m = Chem.MolFromMolFile(str(ui.table_widget.Tab1.worker.settings.InputFilesPaths[isomerindex]).split('.sdf')[0] + '.sdf',removeHs=False)

        Chem.Compute2DCoords(m)

        drawer = rdMolDraw2D.MolDraw2DSVG(300, 300)

        drawer.DrawMolecule(m, highlightAtoms=atom, highlightAtomColors=highlight)

        drawer.FinishDrawing()

        svg = drawer.GetDrawingText().replace('svg:', '')

        svg_bytes = bytearray(svg, encoding='utf-8')

        self.image.renderer().load(svg_bytes)

        self.image.setGeometry(QtCore.QRect(0, 50, 300, 300))

        ui.update()

        # f = open("f.svg", "w+")

        # f.write(str(svg))


class ConformerTab(QtWidgets.QWidget):
    def __init__(self):

        super(ConformerTab, self).__init__()

        self.layout = QtWidgets.QGridLayout()

        self.setLayout(self.layout)

        self.IsomerSelect = QtWidgets.QComboBox(self)

        self.Isomers = ui.table_widget.Tab1.worker.Isomers

        self.IsomerSelect.addItems(self.Isomer_number())

        self.conformertable = QtWidgets.QTableWidget(self)

        self.layout.addWidget(self.conformertable, 0, 0)

        self.conformertable.setColumnCount(3)

        self.conformertable.setHorizontalHeaderLabels(["Conformer", "Energy", "Population"])

        self.IsomerSelect.currentIndexChanged.connect(self.populate_table)

        self.layout.setContentsMargins(0, 50, 0, 0)

        self.conffigure = Figure()

        self.confcanvas = FigureCanvas(self.conffigure)

        self.conformertable.itemSelectionChanged.connect(self.plot_conformers)

        self.layout.addWidget(self.confcanvas, 0, 1)

        self.confcanvas.mpl_connect('button_press_event', self.selectpoint)

    def selectpoint(self, event):

        self.xpos = event.xdata

        self.ypos = event.ydata

        # find the cloest point to the click

        coords = self.conffig.transData.transform((self.xpos, self.ypos))

        coordinates = np.array(self.conffig.transData.transform(list(zip(self.energies, self.populations))))

        mindis = np.argmin((coordinates[:, 0] - coords[0]) ** 2 + (coordinates[:, 1] - coords[1]) ** 2)

        self.conformertable.selectRow(mindis)

    def Isomer_number(self):

        Isomer_list = []

        for c, i in enumerate(self.Isomers):
            Isomer_list.append("Isomer " + str(c + 1))

        return Isomer_list

    def populate_table(self):

        c = 0

        self.isomerindex = int(str(self.IsomerSelect.currentText())[-1]) - 1

        self.conformertable.setRowCount(0)

        self.conformertable.setRowCount(len(self.Isomers[self.isomerindex].Energies))

        for energy, population in zip(self.Isomers[self.isomerindex].Energies,
                                      self.Isomers[self.isomerindex].Populations):
            self.conformertable.setItem(c, 0, QtWidgets.QTableWidgetItem(str(c)))

            self.conformertable.setItem(c, 1, QtWidgets.QTableWidgetItem(str(energy)))

            self.conformertable.setItem(c, 2, QtWidgets.QTableWidgetItem(str(population)))

            c += 1

        self.conformertable.selectRow(0)

        self.plot_conformers()

    def plot_conformers(self):

        self.conffigure.clear()

        self.conffig = self.conffigure.add_subplot(111)

        self.isomerindex = int(str(self.IsomerSelect.currentText())[-1]) - 1

        self.energies = np.array(self.Isomers[self.isomerindex].Energies)

        self.populations = np.array(self.Isomers[self.isomerindex].Populations)

        s = np.argsort(self.energies)

        self.conffig.plot(self.energies[s], self.populations[s],color = "C1")

        self.conffig.plot(self.energies, self.populations, 'o', color='C0', alpha=0.5)

        self.conffig.set_xlabel("Energy (Kcal)")

        self.conffig.set_ylabel("Population")

        selected_conformer = int(self.conformertable.currentRow())

        E = self.energies[selected_conformer]

        P = self.populations[selected_conformer]

        self.conffig.plot(E, P, 'o', color='red', alpha=0.75)

        self.conffig.plot([0, E], [P, P], color='red', alpha=0.75)

        self.conffig.plot([E, E], [0, P], color='red', alpha=0.75)

        self.confcanvas.draw()


class PyDP4WorkerObject(QtCore.QObject):

    finished = QtCore.pyqtSignal()

    def runPyDP4(self):

        launchdir = Path.cwd()

        print(ui.table_widget.Tab1.settings.OutputFolder)
        os.chdir(ui.table_widget.Tab1.settings.OutputFolder)

        self.NMRData, self.Isomers, self.settings, self.DP4Data,self.DP5Data = PyDP4.main(ui.table_widget.Tab1.settings)
        os.chdir(launchdir)
        self.finished.emit()


class WriteStream(object):
    def __init__(self, queue):
        self.queue = queue

    def write(self, text):
        self.queue.put(text)


class MyReceiver(QtCore.QObject):
    mysignal = QtCore.pyqtSignal(str)

    def __init__(self, queue, *args, **kwargs):
        QtCore.QObject.__init__(self, *args, **kwargs)
        self.queue = queue

    @QtCore.pyqtSlot()
    def run(self):
        while True:
            text = self.queue.get()
            self.mysignal.emit(text)


def lorentzian(p, w, p0, A):
    x = (p0 - p) / (w / 2)
    L = A / (1 + x ** 2)

    return L


def ReadParamFile(f, t):
    infile = open(f, 'r')
    inp = infile.readlines()
    infile.close()

    if t not in inp[0]:
        print("Wrong parameter file type, exiting...")
        quit()

    if t == 'm':
        Cmeans = [float(x) for x in inp[1].split(',')]
        Cstdevs = [float(x) for x in inp[2].split(',')]
        Hmeans = [float(x) for x in inp[3].split(',')]
        Hstdevs = [float(x) for x in inp[4].split(',')]

        return Cmeans, Cstdevs, Hmeans, Hstdevs


q = queue.Queue()

sys.stdout = WriteStream(q)

app = QtWidgets.QApplication(sys.argv)

ui = Window()

ui.show()

thread = QtCore.QThread()
my_receiver = MyReceiver(q)
my_receiver.mysignal.connect(ui.table_widget.Tab1.append_text)
my_receiver.moveToThread(thread)
thread.started.connect(my_receiver.run)
thread.start()

sys.exit(app.exec_())





from keras.layers import Layer

from keras import activations
from keras import initializers
from keras import regularizers
from keras import constraints
from keras.layers import Lambda

import tensorflow as tf
import keras.backend as K

class MessageLayer(Layer):
    """ Implements the matrix multiplication message functions from Gilmer
    2017. This could probably be implemented as a series of other layers, but
    this is more convenient.

    """

    def __init__(self, dropout=0., reducer=None, **kwargs):
        """ 

        dropout : float between 0 and 1
            Whether to apply dropout to individual messages before they are
            reduced to each incoming atom.

        reducer : ['sum', 'mean', 'max', or 'min']
            How to collect incoming messages for each atom. In this library,
            I'm careful to only have messages be a function of the sending
            atom, so we can sort the connectivity matrix by recieving atom.
            That lets us use the `segment_*` methods from tensorflow, instead
            of the `unsorted_segment_*` methods.

        """

        self.dropout = dropout
        self.reducer = reducer

        reducer_dict = {
            None: tf.segment_sum,
            'sum': tf.segment_sum,
            'mean': tf.segment_mean,
            'max': tf.segment_max,
            'min': tf.segment_min
        }

        self._reducer = reducer_dict[reducer]

        super(MessageLayer, self).__init__(**kwargs)
    

    def call(self, inputs, training=None):
        """ Perform a single message passing step, returing the summed messages
        for each recieving atom.

        Inputs are [atom_matrix, bond_matrix, connectivity_matrix]

        atom_matrix : (num_atoms_in_batch, d)
            The input matrix of current hidden states for each atom

        bond_matrix : (num_bonds_in_batch, d, d)
            A matrix of current edge features, with each edge represented as a
            (dxd) matrix.

        connectivity : (num_bonds_in_batch, 2)
            A matrix of (a_i, a_j) pairs that indicates the bond in bond_matrix
            connecting atom_matrix[a_j] to atom_matrix[a_i].
            The first entry indicates the recieving atom.

        """

        atom_matrix, bond_matrix, connectivity = inputs

        # Gather the atom matrix so that each reciever node corresponds with
        # the given bond_matrix entry
        atom_gathered = tf.gather(atom_matrix, connectivity[:, 1])

        # Multiply the bond matrices by the gathered atom matrices
        messages = K.batch_dot(bond_matrix, atom_gathered)

        # Add dropout on a message-by-message basis if desired
        def add_dropout():
            if 0. < self.dropout < 1.:
                return K.dropout(messages, self.dropout)
            else:
                return messages

        dropout_messages = K.in_train_phase(
            add_dropout(), messages, training=training)

        # Sum each message along the (sorted) reciever nodes
        summed_message = self._reducer(dropout_messages, connectivity[:, 0])

        return summed_message


    def compute_output_shape(self, input_shape):
        """ Computes the shape of the output, which should be the same
        dimension as the first input, that atom hidden state """
        
        assert input_shape and len(input_shape) == 3
        assert input_shape[0][-1]  # atom hidden state dimension must be specified
        return input_shape[0]


    def get_config(self):
        config = {
            'dropout': self.dropout,
            'reducer': self.reducer,
        }
        base_config = super(MessageLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class GatherAtomToBond(Layer):
    """ Reshapes the atom matrix (num_atoms_in_batch, d) to the bond matrix
    (num_bonds_in_batch, d) by reindexing according to which atom is involved
    in each bond.

    index : 0 or 1
        whether to gather the sending atoms (1) or recieving atoms (0) for each
        bond.

    """

    def __init__(self, index, **kwargs):
        self.index = index
        super(GatherAtomToBond, self).__init__(**kwargs)
    
    def call(self, inputs):
        atom_matrix, connectivity = inputs
        return  tf.gather(atom_matrix, connectivity[:, self.index])

    def compute_output_shape(self, input_shape):
        """ Computes the shape of the output,
        which should be the shape of the atom matrix with the length
        of the bond matrix """
        
        assert input_shape and len(input_shape) == 2
        return input_shape[0]
        
    def get_config(self):
        config = {
            'index': self.index,
        }
        base_config = super(GatherAtomToBond, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

class Reducer(Layer):
    """ Superclass for reducing methods. 
    
    reducer : ['sum', 'mean', 'max', or 'min']
        How to collect elements for each atom or molecule. In this library,
        I'm careful to only have messages be a function of the sending
        atom, so we can sort the connectivity matrix by recieving atom.
        That lets us use the `segment_*` methods from tensorflow, instead
        of the `unsorted_segment_*` methods.

    """

    def __init__(self, reducer=None, **kwargs):

        self.reducer = reducer

        reducer_dict = {
            None: tf.math.segment_sum,
            'sum': tf.math.segment_sum,
            'unsorted_sum': tf.math.unsorted_segment_sum,
            'mean': tf.math.segment_mean,
            'unsorted_mean': tf.math.unsorted_segment_mean,
            'max': tf.math.segment_max,
            'min': tf.math.segment_min
        }

        self._reducer = reducer_dict[reducer]

        super(Reducer, self).__init__(**kwargs)

    def get_config(self):
        config = {'reducer': self.reducer}
        base_config = super(Reducer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        assert input_shape and len(input_shape) == 2
        # Output shape is (n_graphs, atom_dim)
        return input_shape[0]


class ReduceAtomToMol(Reducer):
    """ Sum over all atoms in each molecule.

    Inputs

    atom_matrix : (num_atoms_in_batch, d)
        atom hidden states for each atom in the batch

    node_graph_indices : (num_atoms_in_batch,)
        A scalar for each atom representing which molecule in the batch the
        atom belongs to. This is generated by the preprocessor class, and
        essentially looks like [0, 0, 0, 1, 1] for a batch with a 3 atom
        molecule and a 2 atom molecule.
    """

    def call(self, inputs):

        atom_matrix, node_graph_indices = inputs
        return self._reducer(atom_matrix, node_graph_indices)

class ReduceBondToPro(Reducer):
    """
    Sums over bonds acoording to bond_index to get target bond properties
    
    Inputs

    bond_matrix : (num_bonds_in_batch, d)
        bond hidden states for each bond in the batch

    bond_index : (bond_atoms_in_batch, )
        A scalar for each bond representing which number in the target property 
        the bond links to. This is generated by the preprocessor class, and
        essentially looks like [-1,-1,0,0,-1,-1,1....]
    """

    def call(self, inputs):
        
        bond_matrix, bond_index, n_pro = inputs
        num_segments = tf.reduce_sum(n_pro)
        return self._reducer(bond_matrix, bond_index, num_segments)

    def compute_output_shape(self, input_shape):
        assert input_shape and len(input_shape) == 3
        return input_shape[0]

class ReduceAtomToPro(Reducer):
    """
    Sums over atoms acoording to atom_index to get target atom properties
    
    Inputs

    atom_matrix : (num_atoms_in_batch, d)
        atom hidden states for each atom in the batch

    atom_index : (atom_atoms_in_batch, )
        A scalar for each atom representing which number in the target property 
        the atom links to. This is generated by the preprocessor class, and
        essentially looks like [-1,-1,0,0,-1,-1,1....]
    """

    def call(self, inputs):
        
        atom_matrix, atom_index, n_pro = inputs
        num_segments = tf.reduce_sum(n_pro)
        return self._reducer(atom_matrix, atom_index, num_segments)

    def compute_output_shape(self, input_shape):
        assert input_shape and len(input_shape) == 3
        return input_shape[0]

class ReduceBondToAtom(Reducer):

    """ Sums over the incoming messages from all sender atoms.

    Inputs: 
    
    bond_matrix : (num_bonds_in_batch, d)
        A matrix of messages coming from each sender atom; one row for each
        bond/edge.

    connectivity : (num_bonds_in_batch, 2)
        A matrix of (a_i, a_j) pairs that indicates the bond in bond_matrix
        connecting atom_matrix[a_j] to atom_matrix[a_i].
        The first entry indicates the recieving atom.

    Again, I'm careful to only have the messages be a function of the sending
    node, such that we can use sorted methods in performing the reduction.

    """

    def call(self, inputs):

        bond_matrix, connectivity = inputs
        return self._reducer(bond_matrix, connectivity[:, 0])


class Squeeze(Layer):
    """ Keras forces inputs to be a vector per entry, so this layer squeezes
    them to a single dimension.

    I.e., node_graph_indices will have shape (num_atoms_in_batch, 1), while its
    easier to work with a vector of shape (num_atoms_in_batch,)
    """

    def call(self, inputs):
        return K.squeeze(inputs, 1)

    def compute_output_shape(self, input_shape):
        return input_shape[:-1]


class Embedding2D(Layer):
    """ Keras typically wants to embed items as a single vector, while for the
    matrix multiplication method of Gilmer 2017 we need a matrix for each bond
    type. This just implements that fairly simple extension of the traditional
    embedding layer.
    """

    def __init__(self, input_dim, output_dim,
                 embeddings_initializer='uniform',
                 embeddings_regularizer=None,
                 embeddings_constraint=None,
                 **kwargs):
        
        super(Embedding2D, self).__init__(**kwargs)

        self.input_dim = input_dim
        self.output_dim = output_dim
        
        self.embeddings_initializer = initializers.get(embeddings_initializer)
        self.embeddings_regularizer = regularizers.get(embeddings_regularizer)
        self.embeddings_constraint = constraints.get(embeddings_constraint)
        
    def build(self, input_shape):

        self.embeddings = self.add_weight(
            shape=(self.input_dim, self.output_dim, self.output_dim),
            initializer=self.embeddings_initializer,
            name='bond_embedding_weights', 
            regularizer=self.embeddings_regularizer,
            constraint=self.embeddings_constraint)
        
        self.built = True
        
    def call(self, inputs):
        return tf.nn.embedding_lookup(self.embeddings, inputs)
        
    def compute_output_shape(self, input_shape):
        return (input_shape[0], self.output_dim, self.output_dim)

    def get_config(self):
        config = {
            'input_dim': self.input_dim,
            'output_dim': self.output_dim,
            'embeddings_initializer':
            initializers.serialize(self.embeddings_initializer),
            'embeddings_regularizer':
            regularizers.serialize(self.embeddings_regularizer),
            'embeddings_constraint':
            constraints.serialize(self.embeddings_constraint),
        }
        base_config = super(Embedding2D, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class EdgeNetwork(Layer):
    """ A layer to embed (bond_type, distance) pairs as a NxN matrix. 
    
    Inputs:
    units : dimension of the output matrix
    bond_classes : number of unique bonds

    First perfoms a 1-hot encoding of the bond_type, then passes the
    (*one_hot_encoding, distance) vector to a dense layer. This is the "Edge
    Network" message described by Gilmer, 2017.
    
    """
    
    def __init__(self, units, bond_classes,
                 activation=None,
                 use_bias=True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        
        super(EdgeNetwork, self).__init__(**kwargs)
        self.units = units
        self.bond_classes = bond_classes
        
        self.activation = activations.get(activation)
        self.use_bias = use_bias
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.activity_regularizer = regularizers.get(activity_regularizer)
        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)
        
    def build(self, input_shape):
        
        self.kernel = self.add_weight(shape=(self.bond_classes + 1, self.units**2),
                                      initializer=self.kernel_initializer,
                                      name='kernel',
                                      regularizer=self.kernel_regularizer,
                                      constraint=self.kernel_constraint)
        if self.use_bias:
            self.bias = self.add_weight(shape=(self.units**2,),
                                        initializer=self.bias_initializer,
                                        name='bias',
                                        regularizer=self.bias_regularizer,
                                        constraint=self.bias_constraint)
        else:
            self.bias = None
            
        self.built = True

    def call(self, inputs):
        
        bond_type, distance = inputs
        bond_type_onehot = tf.one_hot(tf.squeeze(bond_type), self.bond_classes)
        stacked_inputs = tf.concat([bond_type_onehot, distance], 1)
        
        output = K.dot(stacked_inputs, self.kernel)
        if self.use_bias:
            output = K.bias_add(output, self.bias, data_format='channels_last')
        if self.activation is not None:
            output = self.activation(output)
            
        output = tf.reshape(output, [-1, self.units, self.units])
        return output
    
    def compute_output_shape(self, input_shape):
        return (input_shape[0][0], self.units, self.units)
    
    def get_config(self):
        config = {
            'units': self.units,
            'bond_classes': self.bond_classes,
            'activation': activations.serialize(self.activation),
            'use_bias': self.use_bias,
            'kernel_initializer': initializers.serialize(self.kernel_initializer),
            'bias_initializer': initializers.serialize(self.bias_initializer),
            'kernel_regularizer': regularizers.serialize(self.kernel_regularizer),
            'bias_regularizer': regularizers.serialize(self.bias_regularizer),
            'activity_regularizer':
                regularizers.serialize(self.activity_regularizer),
            'kernel_constraint': constraints.serialize(self.kernel_constraint),
            'bias_constraint': constraints.serialize(self.bias_constraint)
        }
        base_config = super(EdgeNetwork, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



import logging
import sys
from importlib import import_module
from pathlib import Path
import dataclasses


logger = logging.getLogger(__name__)


def dft_calculations(mols, workflow, config):
    """
    arguments:
    - mols: list of Molecule objects, handled by the Molecules container
    - workflow: workflow parameters
    - config: dft config
    """

    try:
        Method = import_module(f'.{config["method"]}', "dp5.dft")
        logger.debug(f"Loaded {config['method']} as DFT method")
    except ModuleNotFoundError:
        logger.critical(f"No DFT module {config['method']} found. Terminating.")
        sys.exit(1)

    try:
        method = Method.DFTMethod(config)
        logger.debug(f"Initialised {config['method']}")
    except AttributeError as e:
        logger.critical(
            f"{config['method']} does not implement DFTMethod. Terminating."
        )
        sys.exit(1)

    if workflow["dft_opt"]:
        logger.info("Getting optimised geometries...")

        atoms, geometries, energies = method.opt(mols)

        for mol, m_atoms, m_geoms, m_energies in zip(mols, atoms, geometries, energies):
            mol.atoms = m_atoms
            mol.conformers = m_geoms
            if not workflow["dft_energies"]:
                mol.energies = m_energies
        # now to update the geometries
        # should define an object which only copies relevant attributes

    if workflow["dft_energies"]:
        logger.info("Getting DFT energies...")
        atoms, geometries, energies = method.energy(mols)

        for mol, m_atoms, m_geoms, m_energies in zip(mols, atoms, geometries, energies):
            mol.atoms = m_atoms
            mol.conformers = m_geoms
            mol.energies = m_energies

    if workflow["dft_nmr"]:
        logger.info("Getting DFT NMR data...")
        atoms, geometries, energies, shieldings, shielding_labels = method.nmr(mols)
        for mol, m_atoms, m_geoms, m_energies, m_shieldings, m_labels in zip(
            mols, atoms, geometries, energies, shieldings, shielding_labels
        ):

            mol.conformer_C_pred, mol.C_labels, mol.conformer_H_pred, mol.H_labels = (
                _shielding_to_shift(
                    m_atoms, m_shieldings, config["c13_tms"], config["h1_tms"]
                )
            )
            mol.shielding_labels = m_labels
            # handle the shifts here

            if not (workflow["dft_energies"] or workflow["dft_opt"]):
                mol.energies = m_energies

    return mols


def _shielding_to_shift(atoms, shieldings, c13_ref, h1_ref):
    C_pred = []
    C_labels = []
    H_pred = []
    H_labels = []
    shieldings_by_atom = zip(*shieldings)
    for i, (atom, shielding) in enumerate(zip(atoms, shieldings_by_atom), start=1):
        if atom == "C":
            shift = [
                (c13_ref - c_sh) / (1 - (c13_ref * (10**-6))) for c_sh in shielding
            ]
            C_pred.append(shift)
            C_labels.append("C%i" % i)
        if atom == "H":
            shift = [(h1_ref - h_sh) / (1 - (h1_ref * (10**-6))) for h_sh in shielding]
            H_pred.append(shift)
            H_labels.append("H%i" % i)
    C_pred = [list(conf) for conf in zip(*C_pred)]
    H_pred = [list(conf) for conf in zip(*H_pred)]

    return C_pred, C_labels, H_pred, H_labels



import copy
import numpy as np
from scipy.stats.distributions import norm
from lmfit import Parameters
from .fid_corrections import minimisation


def edge_removal(total_spectral_ydata):

    if total_spectral_ydata[0] > 0:
        i = 0
        while total_spectral_ydata[i] > 0:
            total_spectral_ydata[i] = 0
            i += 1
    else:
        i = 0
        while total_spectral_ydata[i] < 0:
            total_spectral_ydata[i] = 0
            i += 1

    if total_spectral_ydata[-1] > 0:
        i = 1
        while total_spectral_ydata[-i] > 0:
            total_spectral_ydata[-i] = 0
            i += 1
    else:
        i = 1
        while total_spectral_ydata[-i] < 0:
            total_spectral_ydata[-i] = 0
            i += 1
    return total_spectral_ydata


def iterative_peak_picking(
    total_spectral_ydata,
    threshold,
    corr_distance,
):

    mu, std = norm.fit(total_spectral_ydata[0:1000])

    picked_peaks = []

    # find all maxima

    maxima = []

    for point in range(1, len(total_spectral_ydata) - 1):
        if (total_spectral_ydata[point] > total_spectral_ydata[point + 1]) & (
            total_spectral_ydata[point] > total_spectral_ydata[point - 1]
        ):
            maxima.append(point)

    # start fitting process

    fit_y = np.zeros(len(total_spectral_ydata))

    while len(maxima) > 0:

        params = Parameters()

        # find peak with greatest amplitude:

        ind1 = np.argmax(total_spectral_ydata[maxima])

        peak = maxima[ind1]

        picked_peaks.append(peak)

        fit_y = minimisation(peak, fit_y, total_spectral_ydata, corr_distance)

        new_maxima = []

        for ind2 in maxima:

            if total_spectral_ydata[ind2] > threshold * std + fit_y[ind2]:

                new_maxima.append(ind2)

        maxima = copy.copy(new_maxima)

    picked_peaks = sorted(picked_peaks)

    return picked_peaks, fit_y



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 17 15:20:18 2014

@author: ke291

Gets called by PyDP4.py if automatic 5-membered cycle corner-flipping is used.
"""

import logging
from math import pi, cos, sin
from scipy.optimize import minimize
from rdkit import Chem, Geometry
from rdkit.Geometry.rdGeometry import Point3D

logger = logging.getLogger(__name__)


def main(f, ring_atoms):
    """
    Find the axis atoms
    Find all the atoms to be rotated

    Rotate it and the substituents to the other side of the plane

    Saves the rotated files, returns nothing.
    Arguments:
    - f: SD File to be read
    - ring_atoms: a list of ring atoms to permute

    Returns:
    none
    """

    mol = Chem.MolFromMolFile(f, removeHs=False)
    conformer = mol.GetConformer()

    # find atoms making up the five-membered ring
    rings = Chem.GetSSSR(mol)
    ring_to_alter = []
    for ring in rings:
        if len(ring_atoms) == 5:
            if all((at in ring) for at in ring_atoms):
                ring_to_alter = ring
                break
        else:
            if len(ring) == 5 and not ring_is_aromatic(mol, ring):
                ring_to_alter = ring
                break

    if not ring_to_alter:
        logger.info("No five membered rings to rotate in %s. Skipping...", f)
        return ""

    # Find the plane of the 5-membered ring and the outlying atom
    norm, d, out_atom_index = find_five_membered_ring_plane(conformer, ring_to_alter)
    out_atom = mol.GetAtomWithIdx(out_atom_index)
    out_atom_pos = conformer.GetAtomPosition(out_atom_index)

    # Find the atoms connected to the outlying atom and sort them
    # as either part of the ring(axis atoms) or as atoms to be rotated
    backbone_atoms = []  # indices!
    rotate_atoms = []  # indices!

    # let's make everything index-centric!

    # this iteration is atom-centric

    for neighbor_atom in out_atom.GetNeighbors():
        nbr_index = neighbor_atom.GetIdx()
        if nbr_index in ring_to_alter:
            backbone_atoms.append(nbr_index)
        else:
            rotate_atoms.append(nbr_index)
            find_sub_atoms(mol, nbr_index, out_atom_index, rotate_atoms)

    # Simple switch to help detect if the atoms are rotated the right way
    angle = find_rot_angle(conformer, *backbone_atoms, out_atom_index, norm)
    rotation_angle = 2 * (0.5 * pi - angle)
    if angle > 0.5 * pi:
        was_obtuse = True
        rotation_angle = 2 * (angle - 0.5 * pi)
    else:
        was_obtuse = False
        rotation_angle = 2 * (0.5 * pi - angle)

    logger.info(
        "Atom "
        + str(out_atom_index)
        + " will be rotated by "
        + str(rotation_angle * 57.3)
        + " degrees"
    )

    # do not confuse positions, angles

    new_angle = find_rot_angle(conformer, *backbone_atoms, out_atom_index, norm)

    if ((new_angle > 0.5 * pi) and was_obtuse) or (
        (new_angle < 0.5 * pi) and not was_obtuse
    ):
        rotation_angle = -rotation_angle

    new_coords = rotate_atom_coords(conformer, out_atom_index, *backbone_atoms, angle)
    conformer.SetAtomPosition(out_atom_index, new_coords)

    rotated_atoms = []
    for atom_index in rotate_atoms:
        if atom_index not in rotated_atoms:
            new_coords = rotate_atom_coords(
                conformer, atom_index, *backbone_atoms, angle
            )
            conformer.SetAtomPosition(atom_index, new_coords)
            rotated_atoms.append(atom_index)
        else:
            logger.info("Atom already rotated, skipping")

    new_id = mol.AddConformer(conformer, assignId=True)
    Chem.MolToMolFile(mol, f[:-4] + "rot.sdf", confId=new_id)
    logger.info("Five membered rings processed in %s.", f)
    return f


def ring_is_aromatic(mol, ring_system):
    """iterates through the rings, confirms all bonds are aromatic
    Arguments:
    - mol(rdkit.Chem.Mol): rdkit Mol object
    - ring_system(tuple or list): a collection containing ring indices. Assumes the order of the atoms follows the ring
    Returns a boolean
    """
    ring_size = len(ring_system)

    for index in range(ring_size):
        i1, i2 = ring_system[index], ring_system[(index + 1) % ring_size]
        if not mol.GetAtomWithIdx(i1).GetIsAromatic():
            return False
        bond = mol.GetBondBetweenAtoms(i1, i2)
        if not bond.GetIsAromatic():
            return False

    return True


def find_plane(atom1: Point3D, atom2: Point3D, atom3: Point3D):

    vector1 = atom2 - atom1
    vector2 = atom3 - atom1

    normal_vector = vector1.CrossProduct(vector2)
    d = atom1.DotProduct(normal_vector)

    return normal_vector, d


def point_plane_dist(norm: Point3D, d: float, atom: Point3D):
    norm = Point3D(*norm)
    a = abs(norm.DotProduct(atom) - d)
    n_length = norm.Length()
    return a / n_length


def plane_error(atoms, a, b, c, d):
    dists = []
    for atom in atoms:
        dists.append(point_plane_dist([a, b, c], d, atom))
    return sum(dists) / len(dists)


def least_squares_plane(atom1, atom2, atom3, atom4):
    [a0, b0, c0], d0 = find_plane(atom1, atom2, atom3)

    f = lambda a: plane_error([atom1, atom2, atom3, atom4], a[0], a[1], a[2], a[3])
    res = minimize(f, (a0, b0, c0, d0), method="nelder-mead")
    plane = list(res.x)

    return Point3D(*plane[:3]), plane[3], f(plane)


def find_five_membered_ring_plane(conformer, atom_ids):
    """returns float, float, outlying atom index"""
    atoms = [conformer.GetAtomPosition(i) for i in atom_ids]
    min_error = 100.0

    for index, atom in enumerate(atom_ids):
        other_atoms = atoms[:index] + atoms[index + 1 :]
        norm, d, error = least_squares_plane(*other_atoms)
        if error < min_error:
            min_error = error
            max_norm = norm
            max_d = d
            out_atom = atom

    return max_norm, max_d, out_atom


# REDO!
def find_sub_atoms(mol, atom_index: int, out_index: int, rotate_atoms: list[int]):
    "Finds neighbourhood of an atom recursively"
    # tread lightly
    atom = mol.GetAtomWithIdx(atom_index)
    for neighbor_atom in atom.GetNeighbors():
        nbr_index = neighbor_atom.GetIdx()
        if (nbr_index not in rotate_atoms) and nbr_index != out_index:
            rotate_atoms.append(nbr_index)
            find_sub_atoms(mol, nbr_index, out_index, rotate_atoms)


def find_rot_angle(
    conformer,
    backbone_index1: int,
    backbone_index2: int,
    out_atom_index: int,
    norm: Point3D,
):
    backbone_atom1 = conformer.GetAtomPosition(backbone_index1)
    backbone_atom2 = conformer.GetAtomPosition(backbone_index2)
    out_atom = conformer.GetAtomPosition(out_atom_index)

    halfway_vec = (backbone_atom1 + backbone_atom2) / 2
    vector = out_atom - halfway_vec
    vangle = vector.AngleTo(norm)
    return vangle


def rotate_atom_coords(
    conformer, atom: int, backbone_index1: int, backbone_index2: int, angle
):
    backbone_atom1 = conformer.GetAtomPosition(backbone_index1)
    backbone_atom2 = conformer.GetAtomPosition(backbone_index2)
    k = backbone_atom1 - backbone_atom2
    k.Normalize()
    v = conformer.GetAtomPosition(atom)
    # Rodrigues formula
    v_rot = (
        v * cos(angle)
        + k.CrossProduct(v) * sin(angle)
        + k * k.DotProduct(v) * (1 - cos(angle))
    )
    return v_rot



from pathlib import Path
import warnings

import numpy as np
import nmrglue as ng


__all__ = ['read_bruker', 'read_jcamp']


def _read_fid(path, reading_func, udic_function):
    """Abstract function for format-specific parsers.
    Arguments:
    - path(str): path to the file/folder
    - reading_func: function that reads file
    - udic_function: function that generates udic
    Returns:
    - nucleus(str): nucleus
    - total_spectral_ydata: complex array containing FID Data
    - uc: nmrglue unit conversion object
    Reads the FID file using the supplied function, returns nucleus for 1D experiments, 
    normalised intensity data, and nmrglue unit conversion object that defines the x-axis.
    """
    dic, total_spectral_ydata = reading_func(path)
    udic0 = udic_function(dic, total_spectral_ydata)
    nucleus = _get_nucleus_from_udic(udic0)
    if nucleus == '1H':
        total_spectral_ydata = preprocess_proton(total_spectral_ydata)
    elif nucleus == '13C':
        total_spectral_ydata = preprocess_carbon(total_spectral_ydata)
    # second generation of udic required after zero-filling
    udic = udic_function(dic, total_spectral_ydata)
    uc = ng.fileiobase.uc_from_udic(udic)
    return nucleus, total_spectral_ydata, uc


def _read_jcamp(file):
    with warnings.catch_warnings(record=True) as warns:
        warnings.filterwarnings(
            "error", category=UserWarning, message='^NTUPLES: multiple')
        dic, total_spectral_ydata = ng.jcampdx.read(file)  # read file

    total_spectral_ydata = total_spectral_ydata[0] + \
        1j * total_spectral_ydata[1]
    total_spectral_ydata = ng.proc_base.ifft_positive(total_spectral_ydata)
    return dic, total_spectral_ydata


def _read_bruker(folder):
    dic, total_spectral_ydata = ng.bruker.read(folder)
    total_spectral_ydata = ng.bruker.remove_digital_filter(
        dic, total_spectral_ydata)
    return dic, total_spectral_ydata


def read_jcamp(file):
    return _read_fid(file, reading_func=_read_jcamp, udic_function=jcamp_guess_udic)


def _get_nucleus_from_udic(udic: dict):
    ndims = udic.get('ndim', 0)
    if ndims == 1:
        nucleus = udic[0]['label']
    else:
        raise NotImplementedError(f"{ndims}-dimensional spectra not supported")
    return nucleus


def read_bruker(folder):
    """Wrapper function for reading bruker dataset folder. Note that the ydata comes normalised"""
    return _read_fid(folder, reading_func=_read_bruker, udic_function=ng.bruker.guess_udic)


def preprocess(total_spectral_ydata: np.array, zero_filling: int) -> np.array:
    """Method-agnostic zero-filling, Fourier transform, and normalisation
    Arguments:
    - total_spectral_ydata: numpy array containing time domain FID
    - zero_filling(int): double length of the time domain this number of times
    Returns:
    - Normalised complex frequency-domain data, highest modulus set to 1
    """
    # Zero-filling – FID time domain doubled 4 times (16 times original length)
    total_spectral_ydata = ng.proc_base.zf_double(
        total_spectral_ydata, zero_filling)
    # Fourier transform into frequency domain
    total_spectral_ydata = ng.proc_base.fft_positive(total_spectral_ydata)

    # Normalise the modulus for the most intense peak to 1
    total_spectral_ydata = normalise_intensities(total_spectral_ydata)

    return total_spectral_ydata


def normalise_intensities(ydata: np.ndarray):
    """Scales array by the largest modulus of the values"""
    return ydata / np.abs(ydata).max()


def preprocess_proton(total_spectral_ydata):
    return preprocess(total_spectral_ydata, zero_filling=4)


def preprocess_carbon(total_spectral_ydata):
    return preprocess(total_spectral_ydata, zero_filling=2)


def jcamp_guess_udic(dic, data):
    """
    Guess parameters of universal dictionary from dic, data pair.
    Parameters
    ----------
    dic : dict
        Dictionary of JCAMP-DX parameters.
    data : ndarray
        Array of NMR data.
    Returns
    -------
    udic : dict
        Universal dictionary of spectral parameters.
    """

    # create an empty universal dictionary
    udic = ng.fileiobase.create_blank_udic(1)

    # update default values (currently only 1D possible)
    # "label"
    try:
        label_value = dic[".OBSERVENUCLEUS"][0].replace("^", "")
        udic[0]["label"] = label_value
    except KeyError:
        # sometimes INSTRUMENTAL PARAMETERS is used:
        try:
            label_value = dic["INSTRUMENTALPARAMETERS"][0].replace("^", "")
            udic[0]["label"] = label_value
        except KeyError:
            pass

    # "obs"
    obs_freq = None
    try:
        obs_freq = float(dic[".OBSERVEFREQUENCY"][0])
        udic[0]["obs"] = obs_freq
    except KeyError:
        pass

    # "size"
    if isinstance(data, list):
        data = data[0]  # if list [R,I]
    if data is not None:
        udic[0]["size"] = len(data)

    # "sw"
    # get firstx, lastx and unit
    firstx, lastx, isppm = ng.jcampdx._find_firstx_lastx(dic)

    # ppm data: convert to Hz
    if isppm:
        if obs_freq:
            firstx = firstx * obs_freq
            lastx = lastx * obs_freq
        else:
            firstx, lastx = (None, None)

    if firstx is not None and lastx is not None:
        udic[0]["sw"] = abs(lastx - firstx)

        # keys not found in standard&required JCAMP-DX keys and thus left default:
        # car, complex, encoding

        udic[0]["car"] = firstx - abs(lastx - firstx) / 2

    return udic


def generalised_lorentzian(
    x: float, mu: float, std: float, v: float, A: float
) -> float:
    """
    Calculate the value of a generalized Lorentzian function at given x values.

    The generalized Lorentzian function is defined as:
    y = A * ((1 - v) / (1 + ((x - mu) / (std / 2)) ** 2) + v * (1 + (((x - mu) / (std / 2)) ** 2) /
    (1 + ((x - mu) / (std / 2)) ** 2 + ((x - mu) / (std / 2)) ** 4))

    Parameters:
    - x (float or array-like): The x values at which to evaluate the function.
    - mu (float): The mean or center of the distribution.
    - std (float): The standard deviation of the distribution.
    - v (float): A parameter controlling the shape of the function.
    - A (float): The amplitude of the function.

    Returns:
    - float or array-like: The value(s) of the generalized Lorentzian function at the given x value(s).
    """
    x1 = (mu - x) / (std / 2)
    y = (1 - v) * (1 / (1 + (x1) ** 2)) + (v) * (
        (1 + ((x1) ** 2) / 2) / (1 + (x1) ** 2 + (x1) ** 4)
    )
    y *= A

    return y


def lorentzian(p, w, p0, A):
    return generalised_lorentzian(p, p0, w, 1, A)


def methyl_protons(mol):

    protons = []

    for atom in mol.GetAtoms():
        if atom.GetAtomicNum() == 6:
            nbrprotons = []
            for nbr in atom.GetNeighbors():
                if nbr.GetAtomicNum() == 1:
                    nbrprotons.append(f"H{nbr.GetIdx()+1}")
            if len(nbrprotons) == 3:
                protons.append(nbrprotons)

    return protons


def labile_protons(mol):
    """
    Counts protons in OH functional groups.

    Assumes those to be exchangeable

    arguments:
        - mol: rdkit Mol object
    returns:
        - count number of labile protons
    """

    count = 0
    for atom in mol.GetAtoms():
        if atom.GetAtomicNum() == 8:
            for neighbor in atom.GetNeighbors():
                if neighbor.GetAtomicNum() == 1:
                    count += 1

    return count


def proton_count(mol) -> int:
    """
    Counts number of hydrogen atoms in a moecule

    arguments:
        - mol: RDKit Mol object
    returns:
        - number of protons (integer)
    """
    return sum([1 for at in mol.GetAtoms() if at.GetAtomicNum() == 1])



from dp5.neural_net.nfp.layers import *
from dp5.neural_net.nfp.models import *

custom_layers = {
    'MessageLayer': MessageLayer,
    'EdgeNetwork': EdgeNetwork,
    'ReduceAtomToMol': ReduceAtomToMol,
    'ReduceBondToAtom': ReduceBondToAtom,
    'GatherAtomToBond': GatherAtomToBond,
    'GRUStep': GRUStep,
    'Embedding2D': Embedding2D,
    'Squeeze': Squeeze,
    'GraphModel': GraphModel,
    'masked_mean_squared_error': masked_mean_squared_error
}



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 17 15:20:18 2014

@author: ke291

Code for diastereomer, tautomer and protomer generation via InChI strings.
This file gets called by PyDP4.py if diastereomer and/or tautomer and/or
protomer generation is used.
"""
from PyDP4 import settings
import sys
import os

try:
    from openbabel.openbabel import OBConversion, OBMol, OBAtomAtomIter, OBMolAtomIter
except ImportError:
    from openbabel import *

import subprocess
from rdkit import Chem
from rdkit.Chem import AllChem

def main(f):

    inchi, aux = GetInchi(f)

    ds_inchis = GenDiastereomers(inchi)
    ds_inchis = [FixTautProtons(f, i, aux) for i in ds_inchis]

    for ds in range(0, len(ds_inchis)):

        print("Isomer " + str(ds) + " inchi = " + ds_inchis[ds])

        Inchi2Struct(ds_inchis[ds], f[:-4] + str(ds+1), aux)
        RestoreNumsSDF(f[:-4] + str(ds+1) + '.sdf', f, aux)


def GetInchiRenumMap(AuxInfo):

    for l in AuxInfo.split('/'):
        if 'N:' in l:
            RenumLayer = l
            break
    amap = [int(x) for x in RenumLayer[2:].split(',')]
    return amap


def FixTautProtons(f, inchi, AuxInfo):

    #Get tautomeric protons and atoms they are connected to from Inchi
    TautProts = GetTautProtons(inchi)
    amap = GetInchiRenumMap(AuxInfo)

    #get the correspondence of the Inchi numbers to the source numbers
    hmap = []
    for taut in TautProts:
        for heavyatom in range(1, len(taut)):
            hmap.append([int(taut[heavyatom]), amap[int(taut[heavyatom])-1]])

    #Read molecule from file
    obconversion = OBConversion()
    obconversion.SetInFormat("sdf")
    obmol = OBMol()
    obconversion.ReadFile(obmol, f)

    Fixprotpos = []
    for heavyatom in hmap:
        atom = obmol.GetAtom(heavyatom[1])
        for nbratom in OBAtomAtomIter(atom):
            if nbratom.GetAtomicNum() == 1:
                Fixprotpos.append(heavyatom[0])
    draftFH = []
    for i in range(0, len(Fixprotpos)):
        if Fixprotpos[i] not in [a[0] for a in draftFH]:
            draftFH.append([Fixprotpos[i], Fixprotpos.count(Fixprotpos[i])])

    fixedlayer = '/f/h'
    for h in draftFH:
        if h[1] == 1:
            fixedlayer = fixedlayer + str(h[0])+'H,'
        else:
            fixedlayer = fixedlayer + str(h[0])+'H' + str(h[1]) + ','

    resinchi = inchi + fixedlayer[:-1]

    return resinchi


#Get H connections from sdf file
def GetHcons(f):
    obconversion = OBConversion()
    obconversion.SetInFormat("sdf")
    obmol = OBMol()
    obconversion.ReadFile(obmol, f)
    Hcons = []
    for atom in OBMolAtomIter(obmol):
        idx = atom.GetIdx()
        anum = atom.GetAtomicNum()
        if anum == 1:
            for NbrAtom in OBAtomAtomIter(atom):
                Hcons.append([idx, NbrAtom.GetIdx()])
    return Hcons


def RestoreNumsSDF(f, fold, AuxInfo):

    #Read molecule from file
    obconversion = OBConversion()
    obconversion.SetInFormat("sdf")
    obmol = OBMol()
    obconversion.ReadFile(obmol, f)
    #Get the atoms Hs are connected to
    oldHcons = GetHcons(fold)
    #translate the H connected atoms to the new numbering system
    amap = GetInchiRenumMap(AuxInfo)
    for i in range(0, len(oldHcons)):
        oldHcons[i][1] = amap.index(oldHcons[i][1])+1

    newHcons = []
    temp = []
    i = 0
    for atom in OBMolAtomIter(obmol):
        idx = atom.GetIdx()
        anum = atom.GetAtomicNum()
        #If atom is hydrogen, check what it is connected to
        if anum == 1:
            for NbrAtom in OBAtomAtomIter(atom):
                newHcons.append([idx, NbrAtom.GetIdx()])
        #Pick the temporary atom
        temp.append(atom)

    for i in range(0, len(newHcons)):
        conatom = newHcons[i][1]
        for b in range(0, len(oldHcons)):
            if conatom == oldHcons[b][1]:
                amap.append(oldHcons[b][0])
                #remove the number, so that it doesn't get added twice
                oldHcons[b][1] = 0

    newmol = OBMol()
    added = []

    for i in range(1, len(amap)+1):
        newn = amap.index(i)
        newmol.AddAtom(temp[newn])
        added.append(newn)

    #Final runthrough to check that all atoms have been added,
    #tautomeric protons can be missed. If tautomeric proton tracking
    #is implemented this can be removed
    for i in range(0, len(temp)):
        if not i in added:
            newmol.AddAtom(temp[i])

    #Restore the bonds
    newmol.ConnectTheDots()
    newmol.PerceiveBondOrders()
    #Write renumbered molecule to file
    obconversion.SetOutFormat("sdf")
    obconversion.WriteFile(newmol, f)


def GetInchi(f):

    print("Getting inchi from file ",f)

    if os.path.sep not in f:
        f = os.path.join(os.getcwd(), f)

    m = Chem.MolFromMolFile(f, removeHs = False)

    m = Chem.AddHs(m)

    idata = Chem.MolToInchiAndAuxInfo(m)

    return idata[0], idata[1]


def Inchi2Struct(inchi, f, aux):


    cwd = os.getcwd()
    fullf = os.path.join(cwd, f)
    infile = open(f + '.inchi', 'w')
    infile.write(inchi)
    infile.close()

    inchi = open(f + '.inchi', "r").read()

    m = AllChem.inchi.MolFromInchi(inchi, sanitize=True, removeHs=False)

    m = AllChem.AddHs(m, addCoords=True)

    AllChem.EmbedMolecule(m)

    save3d = Chem.SDWriter(fullf + '.sdf')

    save3d.write(m)


def GetTautProtons(inchi):
    #get the tautomer layer and pickup the data
    layers = inchi.split('/')

    for l in layers:
        if 'h' in l:
            ProtLayer = l
    ProtList = list(ProtLayer)
    starts = []
    ends = []
    for i in range(0, len(ProtList)):
        if ProtList[i] == '(':
            starts.append(i)
        if ProtList[i] == ')':
            ends.append(i)
    TautProts = []
    for i in range(0, len(starts)):
        TautProts.append((ProtLayer[starts[i]+1:ends[i]]).split(','))

    return TautProts


def GenSelectDiastereomers(structf, atoms):

    f = structf

    if (f[-4:] != '.sdf'):
        f += '.sdf'

    inchi, aux = GetInchi(f)
    amap = GetInchiRenumMap(aux)

    translated_atoms = []
    for atom in atoms:
        translated_atoms.append(amap.index(atom)+1)

    ds_inchis = GenSelectDSInchis(inchi, translated_atoms)
    ds_inchis = [FixTautProtons(f, i, aux) for i in ds_inchis]
    filenames = []
    for ds in range(0, len(ds_inchis)):
        Inchi2Struct(ds_inchis[ds], f[:-4] + str(ds + 1), aux)
        RestoreNumsSDF(f[:-4] + str(ds + 1) + '.sdf', f, aux)
        filenames.append(f[:-4] + str(ds + 1))

    return filenames


def GenSelectDSInchis(inchi, atoms):
    #Inchis of all diastereomers, including the parent structure
    resinchis = []

    #get the number of potential diastereomers
    layers = inchi.decode().split('/')
    for l in layers:
        if 't' in l:
            slayer = l
            sc = l[1:].decode().split(',')

    ignore = []
    for i in range(0, len(sc)):
        if not int(sc[i][:-1]) in atoms:
            ignore.append(sc[i])
    sc = [x for x in sc if x not in ignore]

    if len(sc) == 0:
        "No stereocentres remaining, no diastereomers will be generated."
        return 0

    numds = 2**(len(sc))
    print("Number of diastereomers to be generated: " + str(numds))
    temps = []
    #Generate inversion patterns - essentially just binary strings
    for i in range(0, numds):
        template = bin(i)[2:].zfill(len(sc))
        temps.append(template)

    #For each 1 in template, invert the corresponding stereocentre
    #and add the resulting diastereomer to the list
    invert = {'+': '-', '-': '+'}

    reslayers = []
    for ds in range(0, numds):
        newds = list(sc)
        for stereocentre in range(0, len(sc)):
            if temps[ds][stereocentre] == '1':
                tlist = list(newds[stereocentre])
                tlist[-1] = invert[tlist[-1]]
                newds[stereocentre] = "".join(tlist)
        newlayer = str(slayer)
        for stereocentre in range(0, len(sc)):
            newlayer = newlayer.replace(sc[stereocentre], newds[stereocentre])
        reslayers.append(newlayer)
    print(reslayers)
    resinchis = []
    for layer in reslayers:
        resinchis.append(inchi.replace(slayer, layer))
    return resinchis


def GenDiastereomers(structf, nS, atoms=[]):
    if len(atoms) > 0:
        return GenSelectDiastereomers(structf, atoms)

    f = structf

    if (f[-4:] != '.sdf'):
        f += '.sdf'

    if nS < 2:
        cwd = os.getcwd()

        fullf = os.path.join(cwd, f)

        shutil.copy(fullf, fullf[:-4] + "0.sdf")

        return [f[:-4] + "0.sdf"]

    inchi, aux = GetInchi(f)


    i,a = GetInchi(f)

    ds_inchis = GenDSInchis(inchi)
    ds_inchis = [FixTautProtons(f, i, aux) for i in ds_inchis]
    filenames = []

    for ds in range(0, len(ds_inchis)):

        print("Isomer " + str(ds) + " inchi = " + ds_inchis[ds])

        Inchi2Struct(ds_inchis[ds], f[:-4] + str(ds + 1), aux)
        RestoreNumsSDF(f[:-4] + str(ds + 1) + '.sdf', f, aux)
        filenames.append(f[:-4] + str(ds + 1))

    return filenames


def GenDSInchis(inchi):

    ilist = list(inchi)
    #Inchis of all diastereomers, including the parent structure
    resinchis = []

    #get the number of potential diastereomers
    numds = 0
    layers = inchi.split('/')
    for l in layers:
        if 't' in l:
            numds = 2**(len(l.translate({ord(i): None for i in 't,1234567890'}))-1)

    if numds == 0:

        raise ValueError("No chiral carbon detected in the input molecule!")

    else:
        print("Number of diastereomers to be generated: " + str(numds))

    #find configuration sites (+ and -)
    bs = ilist.index('t')
    es = ilist[bs:].index('/')
    spos = []
    for s in range(bs, bs+es):
        if ilist[s] == '+' or ilist[s] == '-':
            spos.append(s)

    temps = []
    #Generate inversion patterns - essentially just binary strings
    for i in range(0, numds):
        template = bin(i)[2:].zfill(len(spos)-1)
        temps.append(template)

    #For each 1 in template, invert the corresponding stereocentre
    #and add the resulting diastereomer to the list
    invert = {'+': '-', '-': '+'}

    for ds in range(0, numds):
        t = list(ilist)
        for stereocentre in range(1, len(spos)):
            if temps[ds][stereocentre-1] == '1':
                t[spos[stereocentre]] = invert[t[spos[stereocentre]]]
        resinchis.append(''.join(t))

    return resinchis



from dp5.nmr_processing.nmr_ai import NMRData

from pathlib import Path
import pickle
import logging

from rdkit import Chem

logger = logging.getLogger(__name__)

data = NMRData(
    ["/home/rk582/DP5/test/NMR_data/Compound1/6011"], "chloroform", Path.cwd()
)

print(f"NMRData has processed carbon FID: {hasattr(data, 'carbon_fid')}")

with open("/home/rk582/DP5/test/cmpd1.preint", "wb") as f:
    pickle.dump(data, f)



# -*- coding: utf-8 -*-
"""
Created on Wed May 27 14:18:37 2015
Updated on July 30 14:18:37 2015
@author: ke291

Equivalent and compact port of DP4.jar to python. The results
produced are essentially equivalent, but not identical due to different
floating point precision used in the Python (53 bits) and Java (32 bits)
implementation.
"""
from scipy import stats
import bisect
import os
import numpy as np
import pickle

# Standard DP4 parameters
meanC = 0.0
meanH = 0.0
stdevC = 2.269372270818724
stdevH = 0.18731058105269952


class DP4data:
    def __init__(self):
        self.Cshifts = []  # Carbon shifts used in DP4 calculation
        self.Cexp = []  # Carbon experimental shifts used in DP4 calculation
        self.Clabels = []  # Carbon atom labels
        self.Hshifts = []  # Proton shifts used in DP4 calculation
        self.Hexp = []  # Proton experimental shifts used in DP4 calculation
        self.Hlabels = []  # Proton atom labels
        self.Cscaled = []  # Internally scaled carbon shifts
        self.Hscaled = []  # Internally scaled proton shifts
        self.Cerrors = []  # Scaled Carbon prediction errors
        self.Herrors = []  # Scaled Proton prediction errors
        self.Cprobs = []  # Scaled carbon prediction error probabilities
        self.Hprobs = []  # Scaled proton prediction error probabilities
        self.CDP4probs = []  # Carbon DP4 probabilities
        self.HDP4probs = []  # Proton DP4 probabilities
        self.DP4probs = []  # combined Carbon and Proton DP4 probabilities
        self.output = str()  # final DP4 output


def ProcessIsomers(DP4data, Isomers):
    # extract calculated and experimental shifts and add to DP4data instance

    # Carbon

    # make sure any shifts with missing peaks are removed from all isomers

    for iso in Isomers:

        DP4data.Cshifts.append([])
        DP4data.Cexp.append([])
        DP4data.Clabels.append([])

        for shift, exp, label in zip(iso.Cshifts, iso.Cexp, iso.Clabels):

            if exp != '':
                DP4data.Cshifts[-1].append(shift)
                DP4data.Cexp[-1].append(exp)
                DP4data.Clabels[-1].append(label)

    # proton
    for iso in Isomers:

        DP4data.Hshifts.append([])
        DP4data.Hexp.append([])
        DP4data.Hlabels.append([])

        for shift, exp, label in zip(iso.Hshifts, iso.Hexp, iso.Hlabels):

            if exp != '':
                DP4data.Hshifts[-1].append(shift)
                DP4data.Hexp[-1].append(exp)
                DP4data.Hlabels[-1].append(label)

    return DP4data


def InternalScaling(DP4data):
    # perform internal scaling process

    # calculate prediction errors

    if len(DP4data.Cexp[0]) > 1:

        for Cshifts, Cexp in zip(DP4data.Cshifts, DP4data.Cexp):
            DP4data.Cscaled.append(ScaleNMR(Cshifts, Cexp))

    else:
        DP4data.Cscaled = DP4data.Cshifts

    for Cscaled, Cexp in zip(DP4data.Cscaled, DP4data.Cexp):
        DP4data.Cerrors.append([Cscaled[i] - Cexp[i] for i in range(0, len(Cscaled))])


    if len(DP4data.Hexp[0]) > 1:

        for Hshifts, Hexp in zip(DP4data.Hshifts, DP4data.Hexp):
            DP4data.Hscaled.append(ScaleNMR(Hshifts, Hexp))

    else:
        DP4data.Hscaled = DP4data.Hshifts

    for Hscaled, Hexp in zip(DP4data.Hscaled, DP4data.Hexp):
        DP4data.Herrors.append([Hscaled[i] - Hexp[i] for i in range(0, len(Hscaled))])

    return DP4data


def ScaleNMR(calcShifts, expShifts):
    slope, intercept, r_value, p_value, std_err = stats.linregress(expShifts,
                                                                   calcShifts)
    scaled = [(x - intercept) / slope for x in calcShifts]

    return scaled


def CalcProbs(DP4data, Settings):
    # calculates probability values for each scaled prediction error value using the chosen statistical model

    if Settings.StatsModel == 'g' or 'm':

        print(Settings.StatsParamFile)

        if Settings.StatsParamFile == 'none':

            print('No stats model provided, using default')

            for errors in DP4data.Cerrors:
                DP4data.Cprobs.append([SingleGausProbability(e, meanC, stdevC) for e in errors])

            for errors in DP4data.Herrors:
                DP4data.Hprobs.append([SingleGausProbability(e, meanH, stdevH) for e in errors])

        else:

            print('Using stats model provided')

            Cmeans, Cstdevs, Hmeans, Hstdevs = ReadParamFile(Settings.StatsParamFile, Settings.StatsModel)

            for errors in DP4data.Cerrors:
                DP4data.Cprobs.append([MultiGausProbability(e, Cmeans, Cstdevs) for e in errors])

            for errors in DP4data.Herrors:
                DP4data.Hprobs.append([MultiGausProbability(e, Hmeans, Hstdevs) for e in errors])

    return DP4data


def SingleGausProbability(error, mean, stdev):
    z = abs((error - mean) / stdev)
    cdp4 = 2 * stats.norm.cdf(-z)

    return cdp4


def MultiGausProbability(error, means, stdevs):
    res = 0

    for mean, stdev in zip(means, stdevs):
        res += stats.norm(mean, stdev).pdf(error)

    return res / len(means)


def ReadParamFile(f, t):
    infile = open(f, 'r')
    inp = infile.readlines()
    infile.close()

    if t not in inp[0]:
        print("Wrong parameter file type, exiting...")
        quit()

    if t == 'm':
        Cmeans = [float(x) for x in inp[1].split(',')]
        Cstdevs = [float(x) for x in inp[2].split(',')]
        Hmeans = [float(x) for x in inp[3].split(',')]
        Hstdevs = [float(x) for x in inp[4].split(',')]

        return Cmeans, Cstdevs, Hmeans, Hstdevs


def CalcDP4(DP4data):
    # Calculate Carbon DP4 probabilities

    for probs in DP4data.Cprobs:

        DP4data.CDP4probs.append(1)

        for p in probs:
            DP4data.CDP4probs[-1] *= p

    # Calculate Proton DP4 probabilities

    for probs in DP4data.Hprobs:

        DP4data.HDP4probs.append(1)

        for p in probs:
            DP4data.HDP4probs[-1] *= p

    # Calculate Combined DP4 probabilities

    for Hp, Cp in zip(DP4data.HDP4probs, DP4data.CDP4probs):
        DP4data.DP4probs.append(Hp * Cp)

    Cs = sum(DP4data.CDP4probs)

    Hs = sum(DP4data.HDP4probs)

    Ts = sum(DP4data.DP4probs)

    DP4data.CDP4probs = [i / Cs for i in DP4data.CDP4probs]

    DP4data.HDP4probs = [i / Hs for i in DP4data.HDP4probs]

    DP4data.DP4probs = [i / Ts for i in DP4data.DP4probs]

    return DP4data


def PrintAssignment(DP4Data):
    isomer = 0

    for Clabels, Cshifts, Cexp, Cscaled in zip(DP4Data.Clabels, DP4Data.Cshifts, DP4Data.Cexp, DP4Data.Cscaled):
        DP4Data.output += ("\n\nAssigned C shifts for isomer " + str(isomer + 1) + ": ")

        PrintNMR(Clabels, Cshifts, Cscaled, Cexp, DP4Data)

        isomer += 1

    isomer = 0

    for Hlabels, Hshifts, Hexp, Hscaled in zip(DP4Data.Hlabels, DP4Data.Hshifts, DP4Data.Hexp, DP4Data.Hscaled):
        DP4Data.output += ("\n\nAssigned H shifts for isomer " + str(isomer + 1) + ": ")

        PrintNMR(Hlabels, Hshifts, Hscaled, Hexp, DP4Data)

        isomer += 1


def PrintNMR(labels, values, scaled, exp, DP4Data):
    s = np.argsort(values)

    svalues = np.array(values)[s]

    slabels = np.array(labels)[s]

    sscaled = np.array(scaled)[s]

    sexp = np.array(exp)[s]

    DP4Data.output += ("\nlabel, calc, corrected, exp, error")

    for i in range(len(labels)):
        DP4Data.output += ("\n" + format(slabels[i], "6s") + ' ' + format(svalues[i], "6.2f") + ' '
                           + format(sscaled[i], "6.2f") + ' ' + format(sexp[i], "6.2f") + ' ' +
                           format(sexp[i] - sscaled[i], "6.2f"))


def MakeOutput(DP4Data, Isomers, Settings):
    # add some info about the calculation

    DP4Data.output += Settings.InputFiles[0] + "\n"

    DP4Data.output += "\n" + "Solvent = " + Settings.Solvent

    DP4Data.output += "\n" + "Force Field = " + Settings.ForceField + "\n"

    if 'o' in Settings.Workflow:
        DP4Data.output += "\n" + "DFT optimisation Functional = " + Settings.oFunctional
        DP4Data.output += "\n" + "DFT optimisation Basis = " + Settings.oBasisSet

    if 'e' in Settings.Workflow:
        DP4Data.output += "\n" + "DFT energy Functional = " + Settings.eFunctional
        DP4Data.output += "\n" + "DFT energy Basis = " + Settings.eBasisSet

    if 'n' in Settings.Workflow:
        DP4Data.output += "\n" + "DFT NMR Functional = " + Settings.nFunctional
        DP4Data.output += "\n" + "DFT NMR Basis = " + Settings.nBasisSet

    if Settings.StatsParamFile != "none":
        DP4Data.output += "\n\nStats model = " + Settings.StatsParamFile

    DP4Data.output += "\n\nNumber of isomers = " + str(len(Isomers))

    c = 1

    for i in Isomers:
        DP4Data.output += "\nNumber of conformers for isomer " + str(c) + " = " + str(len(i.Conformers))

        c += 1

    PrintAssignment(DP4Data)

    DP4Data.output += ("\n\nResults of DP4 using Proton: ")

    for i, p in enumerate(DP4Data.HDP4probs):
        DP4Data.output += ("\nIsomer " + str(i + 1) + ": " + format(p * 100, "4.1f") + "%")

    DP4Data.output += ("\n\nResults of DP4 using Carbon: ")

    for i, p in enumerate(DP4Data.CDP4probs):
        DP4Data.output += ("\nIsomer " + str(i + 1) + ": " + format(p * 100, "4.1f") + "%")

    DP4Data.output += ("\n\nResults of DP4: ")

    for i, p in enumerate(DP4Data.DP4probs):
        DP4Data.output += ("\nIsomer " + str(i + 1) + ": " + format(p * 100, "4.1f") + "%")

    print("number of c protons = " + str(len(Isomers[0].Hlabels)))
    print("number of c carbons = " + str(len(Isomers[0].Clabels)))

    print("number of e protons = " + str(len(DP4Data.Hexp[0])))
    print("number of e carbons = " + str(len(DP4Data.Cexp[0])))

    print(DP4Data.output)

    save_results(DP4Data, Settings)

    if Settings.OutputFolder == '':

        out = open(str(os.getcwd()) + "/" + str(Settings.InputFiles[0] + "NMR.dp4"), "w+")

    else:

        out = open(os.path.join(Settings.OutputFolder, str(Settings.InputFiles[0] + "NMR.dp4")), "w+")


    out.write(DP4Data.output)

    out.close()

    return DP4Data

def save_results(DP4Data:DP4data, settings):
        data_dic = {"Cshifts": DP4Data.Cshifts,
                    "Cexp": DP4Data.Cexp,
                    "Clabels": DP4Data.Clabels,
                    "Hshifts": DP4Data.Hshifts,
                    "Hexp": DP4Data.Hexp,
                    "Hlabels": DP4Data.Hlabels,
                    "Cscaled": DP4Data.Cscaled,
                    "Hscaled": DP4Data.Hscaled,
                    "Cerrors": DP4Data.Cerrors,
                    "Herrors": DP4Data.Herrors,
                    "Cprobs": DP4Data.Cprobs,
                    "Hprobs": DP4Data.Hprobs,
                    "CDP4probs": DP4Data.CDP4probs,
                    "HDP4probs": DP4Data.HDP4probs,
                    "DP4probs": DP4Data.DP4probs
                    }
        new_folder = os.path.join(settings.OutputFolder, "dp4")
        try:
            os.mkdir(new_folder)
        except OSError:
            pass
        data_path = os.path.join(new_folder, "data_dic.p")
        with open(data_path, "wb") as path:
            pickle.dump(data_dic, path)
